from __future__ import annotations

from typing import Any

import numpy as np
import pandas as pd

from ..dataset import (
    AdditionalTreatmentRole,
    Dataset,
    ExperimentData,
    StratificationRole,
    TreatmentRole,
)
from ..dataset.roles import ConstGroupRole
from ..executor import Calculator
from ..utils import ExperimentDataEnum


class AASplitter(Calculator):
    def __init__(
        self,
        control_size: float = 0.5,
        random_state: int | None = None,
        sample_size: float | None = None,
        constant_key: bool = True,
        save_groups: bool = True,
        groups_sizes: list[float] | None = None,
        key: Any = "",
    ):
        self.control_size = control_size
        self.random_state = random_state
        self._key = key
        self.constant_key = constant_key
        self.save_groups = save_groups
        self.sample_size = sample_size
        self.groups_sizes = groups_sizes
        super().__init__(key)

    def _generate_params_hash(self):
        hash_parts: list[str] = []
        if self.control_size != 0.5:
            hash_parts.append(f"cs {self.control_size}")
        if self.random_state is not None:
            hash_parts.append(f"rs {self.random_state}")
        if self.groups_sizes is not None:
            hash_parts.append(f"gs {self.groups_sizes}")
        self._params_hash = "|".join(hash_parts)

    def init_from_hash(self, params_hash: str):
        hash_parts: list[str] = params_hash.split("|")
        for hash_part in hash_parts:
            if hash_part.startswith("cs"):
                self.control_size = float(hash_part[hash_part.rfind(" ") + 1 :])
            elif hash_part.startswith("rs"):
                self.random_state = int(hash_part[hash_part.rfind(" ") + 1 :])
            elif hash_part.startswith("gs"):
                self.groups_sizes = []
                groups_sizes = (
                    hash_part[hash_part.find(" ") + 1 :].strip("[]").split(",")
                )
                self.groups_sizes = [float(gs) for gs in groups_sizes]
        self._generate_id()

    @property
    def key(self) -> Any:
        return self._key

    @key.setter
    def key(self, value: Any):
        if not self.constant_key:
            self._key = value
            self._generate_id()

    def _set_value(self, data: ExperimentData, value, key=None) -> ExperimentData:
        data = data.set_value(
            ExperimentDataEnum.additional_fields,
            self._id,
            value,
            role=AdditionalTreatmentRole(),
        )

        if self.save_groups:
            data.groups[self.id] = {
                group: data.ds.loc[group_data.index]
                for group, group_data in data.additional_fields.groupby(self.id)
            }
        return data

    @staticmethod
    def _inner_function(
        data: Dataset,
        random_state: int | None = None,
        control_size: float = 0.5,
        groups_sizes: list[float] | None = None,
        sample_size: float | None = None,
        const_group_field: str | None = None,
        **kwargs,
    ) -> list[str]:
        sample_size = 1.0 if sample_size is None else sample_size
        control_indexes = []
        if const_group_field:
            const_data = dict(data.groupby(const_group_field))
            control_data = const_data.get("control")
            if control_data is not None:
                control_indexes = list(control_data.index)
            const_size = sum(len(cd) for cd in const_data.values())
            control_size = (
                0
                if len(data) <= const_size
                else (len(data) * control_size - len(const_data["control"]))
                / (len(data) - const_size)
            )
            # control_size = len(data) * control_size
        experiment_data = (
            data[data[const_group_field].isna()] if const_group_field else data
        )
        experiment_data_index = experiment_data.sample(
            frac=sample_size, random_state=random_state
        ).index
        addition_indexes = list(experiment_data_index)
        edges = []
        if groups_sizes:
            if sum(groups_sizes) != 1:
                raise ValueError("Groups sizes must sum to 1")
            for group_size in groups_sizes:
                size = int(len(addition_indexes) * group_size) + (
                    0 if not edges else edges[-1]
                )
                size = min(size, len(addition_indexes))
                if size not in edges:
                    edges += [size]
        else:
            edges = [int(len(addition_indexes) * control_size), len(addition_indexes)]
        control_indexes += addition_indexes[: edges[0]]
        test_indexes = [
            addition_indexes[edges[i - 1] : edges[i]] for i in range(1, len(edges))
        ]

        split_series = pd.Series(
            np.ones(data.data.shape[0], dtype="int"), index=data.data.index
        )
        split_series[control_indexes] -= 1
        for i, test_index in enumerate(test_indexes):
            split_series[test_index] += i

        label_map = {0: "control"}
        label_map.update({i: f"test_{i}" for i in range(1, len(edges))})
        split_series = split_series.map(label_map)

        return split_series.to_list()

    def execute(self, data: ExperimentData) -> ExperimentData:
        const_group_fields = data.ds.search_columns(ConstGroupRole())
        const_group_fields = (
            const_group_fields[0] if len(const_group_fields) > 0 else None
        )
        result = self.calc(
            data.ds,
            random_state=self.random_state,
            control_size=self.control_size,
            sample_size=self.sample_size,
            const_group_field=const_group_fields,
            groups_sizes=self.groups_sizes,
        )
        return self._set_value(
            data,
            result,
        )


class AASplitterWithStratification(AASplitter):
    @staticmethod
    def _inner_function(
        data: Dataset,
        random_state: int | None = None,
        control_size: float = 0.5,
        grouping_fields=None,
        **kwargs,
    ) -> list[str] | Dataset:
        if not grouping_fields:
            return AASplitter._inner_function(
                data, random_state, control_size, **kwargs
            )

        result = {"split": []}
        index = []
        for group, group_data in data.groupby(grouping_fields):
            result["split"].extend(
                AASplitter._inner_function(group_data, random_state, control_size)
            )
            index.extend(list(group_data.index))
        return Dataset.from_dict(result, index=index, roles={"split": TreatmentRole()})

    def execute(self, data: ExperimentData) -> ExperimentData:
        grouping_fields = data.ds.search_columns(StratificationRole())
        result = self.calc(
            data.ds,
            random_state=self.random_state,
            control_size=self.control_size,
            grouping_fields=grouping_fields,
            groups_sizes=self.groups_sizes,
        )
        if isinstance(result, Dataset):
            result = result.replace_roles({"split": AdditionalTreatmentRole()})
        return self._set_value(data, result)


#
# class AASplitterWithStratification(AASplitter):
#     def __init__(
#         self,
#         control_size: float = 0.5,
#         random_state: Optional[int] = None,
# #         key: Any = "",
#     ):
#         super().__init__(control_size, random_state,  key)
#
#     def calc(self, data: Dataset):
#         stratification_columns = data.get_columns_by_roles(StratificationRole())
#
#         groups = data.groupby(stratification_columns)
#         result = Dataset._create_empty()
#         for _, gd in groups:
#             ged = ExperimentData(gd)
#             ged = super().execute(ged)
#
#             result = ged if result is None else result.append(ged)
#         return result["group"]


# As idea
# class SplitterAAMulti(ExperimentMulti):
#     def execute(self, data):
#         raise NotImplementedError



from __future__ import annotations

from typing import Any, Sequence


class Adapter:
    @staticmethod
    def to_list(data: Any) -> list:
        if data is None:
            return []
        if isinstance(data, str):
            return [data]
        return list(data) if isinstance(data, Sequence) else [data]

    @staticmethod
    def list_to_single(data: list) -> Any:
        if isinstance(data, list):
            if len(data) == 0:
                return None
            elif len(data) == 1:
                return data[0]
            else:
                raise ValueError("Only a list of a single item can be accepted")



from .analyzers.aa import OneAAStatAnalyzer
from .comparators import Chi2Test, GroupDifference, GroupSizes, KSTest, TTest
from .dataset import TargetRole, TreatmentRole
from .experiments.base import Experiment, OnRoleExperiment
from .ui.base import ExperimentShell
from .ui.homo import HomoOutput

HOMOGENEITY_TEST = Experiment(
    executors=[
        OnRoleExperiment(
            executors=[
                GroupSizes(grouping_role=TreatmentRole(), compare_by="groups"),
                GroupDifference(grouping_role=TreatmentRole(), compare_by="groups"),
                TTest(grouping_role=TreatmentRole(), compare_by="groups"),
                KSTest(grouping_role=TreatmentRole(), compare_by="groups"),
                Chi2Test(grouping_role=TreatmentRole(), compare_by="groups"),
            ],
            role=TargetRole(),
        ),
        OneAAStatAnalyzer(),
    ]
)


class HomogeneityTest(ExperimentShell):
    """A class for conducting homogeneity tests between the groups.

    This class provides functionality to test whether treatment and control groups are
    homogeneous across target variables using multiple statistical tests including t-test,
    Kolmogorov-Smirnov test, and chi-square test.

    The class runs the following analyses:
        - Group size comparisons
        - Group differences
        - T-test for continuous variables
        - KS-test for distribution comparisons
        - Chi-square test for categorical variables
        - AA statistics analysis

    Examples
    --------
    .. code-block:: python

        # Basic homogeneity test
        homo_test = HomogeneityTest()
        results = homo_test.execute(data)

        # Accessing specific test results
        homo_test = HomogeneityTest()
        results = homo_test.execute(data)
        output = results.resume

        # Running test on dataset with roles
        from hypex.dataset import Dataset, TargetRole, TreatmentRole
        ds = Dataset(
            roles={
                'treatment': TreatmentRole(),
                'outcome': TargetRole()
            },
            data=df
        )
        homo_test = HomogeneityTest()
        results = homo_test.execute(ds)
    """

    def __init__(self):
        """Initialize HomogeneityTest with default experiment and output configurations."""
        super().__init__(
            experiment=HOMOGENEITY_TEST,
            output=HomoOutput(),
        )



from __future__ import annotations

from itertools import product
from typing import Any, Sequence

from tqdm import tqdm

from ..dataset import ABCRole, Dataset, ExperimentData, GroupingRole
from ..executor import Executor, IfExecutor
from ..reporters import DatasetReporter, Reporter
from ..utils.enums import ExperimentDataEnum
from .base import Experiment


class ExperimentWithReporter(Experiment):
    def __init__(
        self,
        executors: Sequence[Executor],
        reporter: Reporter,
        transformer: bool | None = None,
        key: str = "",
    ):
        super().__init__(executors, transformer, key)
        self.reporter = reporter

    def one_iteration(
        self, data: ExperimentData, key: str = "", set_key_as_index: bool = False
    ):
        t_data = ExperimentData(data.ds)
        self.key = key
        t_data = super().execute(t_data)
        result = self.reporter.report(t_data)
        if set_key_as_index:
            result.index = [key]
        return result

    def _set_result(
        self, data: ExperimentData, result: list[Dataset], reset_index: bool = True
    ):
        result = (
            result[0].append(result[1:], reset_index=reset_index)
            if len(result) > 1
            else result[0]
        )
        return self._set_value(data, result)


class CycledExperiment(ExperimentWithReporter):
    def __init__(
        self,
        executors: list[Executor],
        reporter: DatasetReporter,
        n_iterations: int,
        transformer: bool | None = None,
        key: str = "",
    ):
        super().__init__(executors, reporter, transformer, key)
        self.n_iterations: int = n_iterations

    def generate_params_hash(self) -> str:
        return f"{self.reporter.__class__.__name__} x {self.n_iterations}"

    def execute(self, data: ExperimentData) -> ExperimentData:
        result: list[Dataset] = [
            self.one_iteration(data, str(i)) for i in tqdm(range(self.n_iterations))
        ]
        return self._set_result(data, result)


class GroupExperiment(ExperimentWithReporter):
    def __init__(
        self,
        executors: Sequence[Executor],
        reporter: Reporter,
        searching_role: ABCRole = GroupingRole(),
        transformer: bool | None = None,
        key: str = "",
    ):
        self.searching_role = searching_role
        super().__init__(executors, reporter, transformer, key)

    def generate_params_hash(self) -> str:
        return f"GroupExperiment: {self.reporter.__class__.__name__}"

    def execute(self, data: ExperimentData) -> ExperimentData:
        group_field = data.ds.search_columns(self.searching_role)
        result: list[Dataset] = [
            self.one_iteration(
                ExperimentData(group_data), str(group[0]), set_key_as_index=True
            )
            for group, group_data in tqdm(data.ds.groupby(group_field))
        ]
        return self._set_result(data, result, reset_index=False)


class ParamsExperiment(ExperimentWithReporter):
    def __init__(
        self,
        executors: Sequence[Executor],
        reporter: DatasetReporter,
        params: dict[type, dict[str, Sequence[Any]]],
        transformer: bool | None = None,
        key: str = "",
    ):
        super().__init__(executors, reporter, transformer, key)
        self._params = params
        self._flat_params: list[dict[type, dict[str, Any]]] = []

    def generate_params_hash(self) -> str:
        return f"ParamsExperiment: {self.reporter.__class__.__name__}"

    def _update_flat_params(self):
        classes = list(self._params)
        param_combinations = [
            list(
                product(
                    *[
                        product([parameter], values)
                        for parameter, values in class_params.items()
                    ]
                )
            )
            for class_params in self._params.values()
        ]
        new_flat_params = [
            {
                classes[i]: dict(param_combination[i])
                for i in range(len(param_combination))
            }
            for param_combination in product(*param_combinations)
        ]
        self._flat_params = new_flat_params

    @property
    def flat_params(self) -> list[dict[type, dict[str, Any]]]:
        return self._flat_params

    @property
    def params(self) -> dict[type, dict[str, Sequence[Any]]]:
        return self._params

    @params.setter
    def params(self, params: dict[type, dict[str, Sequence[Any]]]):
        self._params = params
        self._update_flat_params()

    def execute(self, data: ExperimentData) -> ExperimentData:
        results = []
        self._update_flat_params()
        for flat_param in tqdm(self._flat_params):
            t_data = ExperimentData(data.ds)
            for executor in self.executors:
                executor.set_params(flat_param)
                t_data = executor.execute(t_data)
            report = self.reporter.report(t_data)
            results.append(report)
        return self._set_result(data, results)


class IfParamsExperiment(ParamsExperiment):
    def __init__(
        self,
        executors: Sequence[Executor],
        reporter: DatasetReporter,
        params: dict[type, dict[str, Sequence[Any]]],
        stopping_criterion: IfExecutor,
        transformer: bool | None = None,
        key: str = "",
    ):
        self.stopping_criterion = stopping_criterion
        super().__init__(executors, reporter, params, transformer, key)

    def execute(self, data: ExperimentData) -> ExperimentData:
        self._update_flat_params()
        for flat_param in tqdm(self._flat_params):
            t_data = ExperimentData(data.ds)
            for executor in self.executors:
                executor.set_params(flat_param)
                t_data = executor.execute(t_data)
            if_result = self.stopping_criterion.execute(t_data)
            if_executor_id = if_result.get_one_id(
                self.stopping_criterion.__class__, ExperimentDataEnum.variables
            )
            if if_result.variables[if_executor_id]["response"]:
                return self._set_result(data, [self.reporter.report(t_data)])
        return data



from typing import Literal


class RoleColumnError(Exception):
    def __init__(self, roles, columns):
        super().__init__(
            "Check your roles. All of them must be names of data columns. \n"
            f"Now roles have {roles} values and columns have {columns} values"
        )


class ConcatDataError(Exception):
    def __init__(self, data_type):
        super().__init__(f"Can only append Dataset to Dataset. Got {data_type}")


class ConcatBackendError(Exception):
    def __init__(self, other_backend, backend):
        super().__init__(
            f"Can only append data with the same backends. Got {other_backend} expected {backend}"
        )


class SpaceError(Exception):
    def __init__(self, space):
        super().__init__(f"{space} is not a valid space")


class NoColumnsError(Exception):
    def __init__(self, role):
        super().__init__(f"No columns found by role {role}")


class NotSuitableFieldError(Exception):
    def __init__(self, field, field_role: Literal["Grouping", "Target", "Baseline"]):
        super().__init__(f"{field_role} field {field} is not suitable for comparison")


class NotFoundInExperimentDataError(Exception):
    def __init__(self, class_: str):
        super().__init__(f"{class_} id is not found in ExperimentData")


class AbstractMethodError(NotImplementedError):
    def __init__(self):
        super().__init__(
            "This method is abstract and will be overridden in derived class."
        )


class DataTypeError(Exception):
    def __init__(self, data_type):
        super().__init__(
            f"Can only perform the operation for Dataset and Dataset. Got {data_type}"
        )


class BackendTypeError(Exception):
    def __init__(self, other_backend, backend):
        super().__init__(
            f"Can only perform the operation with the same backends. Got {other_backend} expected {backend}"
        )


class MergeOnError(Exception):
    def __init__(self, on):
        super().__init__(f"Can only merge on one of the columns data. Got {on}")


class NoRequiredArgumentError(Exception):
    def __init__(self, argument_name):
        super().__init__(f"The required argument {argument_name} has not been passed.")


class NoneArgumentError(Exception):
    def __init__(self, arg, process):
        super().__init__(f"Argument {arg} is None in process {process}.")


class InvalidArgumentError(Exception):
    def __init__(self, arg, possible_type):
        super().__init__(
            f"Invalid type for argument {arg}, possible type is is {possible_type}."
        )


class PairsNotFoundError(Exception):
    def __init__(self):
        super().__init__(
            "Pairs are not found. Check your input data and try execute preprocessing pipeline before matching estimation."
        )



from __future__ import annotations

from typing import Any

from ..dataset import Dataset, ExperimentData
from ..experiments.base import Experiment
from ..reporters import Reporter
from ..utils import ID_SPLIT_SYMBOL
from ..utils.enums import RenameEnum


class Output:
    """A class for handling experiment output reporting and formatting.

    This class manages the reporting and formatting of experiment results, allowing for both
    a primary resume report and additional custom reports.

    Attributes:
        resume (Dataset): The main summary report of the experiment results.
        _experiment_data (ExperimentData): Internal storage of the experiment data.

    Args:
        resume_reporter (Reporter): The main reporter that generates the resume output.
        additional_reporters (Optional[Dict[str, Reporter]]): Dictionary mapping attribute
            names to additional reporters for custom reporting. Defaults to None.

    Examples
    --------
    .. code-block:: python

        # Basic usage with just a resume reporter
        from my_reporters import MyResumeReporter
        output = Output(resume_reporter=MyResumeReporter())
        output.extract(experiment_data)
        print(output.resume)

        # Using additional custom reporters
        from my_reporters import StatsReporter, PlotReporter
        additional = {
            'statistics': StatsReporter(),
            'plots': PlotReporter()
        }
        output = Output(
            resume_reporter=MyResumeReporter(),
            additional_reporters=additional
        )
        output.extract(experiment_data)
        print(output.statistics)  # Access additional report
        print(output.plots)  # Access additional report
    """

    resume: Dataset
    _experiment_data: ExperimentData

    def __init__(
        self,
        resume_reporter: Reporter,
        additional_reporters: dict[str, Reporter] | None = None,
    ):
        self.resume_reporter = resume_reporter
        self.additional_reporters = additional_reporters or {}

    def _extract_by_reporters(self, experiment_data: ExperimentData):
        """Extracts reports from all configured reporters.

        Args:
            experiment_data (ExperimentData): The experiment data to generate reports from.
        """
        self.resume = self.resume_reporter.report(experiment_data)
        for attribute, reporter in self.additional_reporters.items():
            setattr(self, attribute, reporter.report(experiment_data))
        self._experiment_data = experiment_data

    @staticmethod
    def _replace_splitters(
        data: Dataset, mode: RenameEnum = RenameEnum.columns
    ) -> Dataset:
        result = data
        if mode in (RenameEnum.all, RenameEnum.columns):
            result = result.rename(
                {c: c.replace(ID_SPLIT_SYMBOL, " ") for c in result.columns}
            )
        if mode in (RenameEnum.all, RenameEnum.index):
            result.index = [i.replace(ID_SPLIT_SYMBOL, " ") for i in result.index]
        return result

    def extract(self, experiment_data: ExperimentData):
        """Extracts and processes all reports from the experiment data.

        Args:
            experiment_data (ExperimentData): The experiment data to generate reports from.

        Examples
        --------
        .. code-block:: python

            output = Output(resume_reporter=MyReporter())
            output.extract(experiment_data)
            print(output.resume)  # Access the main report
        """
        self._extract_by_reporters(experiment_data)


class ExperimentShell:
    """Base class for experiment execution with configurable output handling.

    This class provides a shell for executing experiments with customizable parameters
    and output formatting. It serves as a base class for specific experiment types
    like A/B tests and A/A tests.

    Args:
        experiment (Experiment): The experiment configuration to execute.
        output (Output): Output handler that defines how results are formatted.
        experiment_params (Optional[Dict[str, Any]], optional): Additional parameters
            to configure the experiment. Defaults to None.

    Examples
    --------
    .. code-block:: python

        # Basic usage with default parameters
        experiment = Experiment([...])  # Configure experiment
        output = Output(resume_reporter=MyReporter())
        shell = ExperimentShell(experiment, output)
        results = shell.execute(data)

        # With custom experiment parameters
        params = {
            "random_state": 42,
            "test_size": 0.3
        }
        shell = ExperimentShell(
            experiment=experiment,
            output=output,
            experiment_params=params
        )
        results = shell.execute(data)
    """

    def __init__(
        self,
        experiment: Experiment,
        output: Output,
        experiment_params: dict[str, Any] | None = None,
    ):
        if experiment_params:
            experiment.set_params(experiment_params)
        self._out = output
        self._experiment = experiment

    @property
    def experiment(self):
        """Gets the configured experiment instance.

        Returns:
            Experiment: The experiment configuration object.
        """
        return self._experiment

    def execute(self, data: Dataset | ExperimentData) -> Output:
        """Executes the experiment on the provided data.

        Runs the configured experiment on the input data and formats the results
        using the configured output handler.

        Args:
            data (Union[Dataset, ExperimentData]): Input data for the experiment.
                Can be either a Dataset or ExperimentData instance.

        Returns:
            Output: Formatted experiment results through the configured output handler.

        Examples
        --------
        .. code-block:: python

            shell = ExperimentShell(experiment, output)
            dataset = Dataset(...)  # Your input data
            results = shell.execute(dataset)
            print(results.resume)  # Access formatted results
        """
        if isinstance(data, Dataset):
            data = ExperimentData(data)
        result_experiment_data = self._experiment.execute(data)
        self._out.extract(result_experiment_data)
        return self._out



from __future__ import annotations

import warnings
from typing import Callable

from scipy.stats import (  # type: ignore
    chi2_contingency,
    ks_2samp,
    mannwhitneyu,
    norm,
    ttest_ind,
)

from ..dataset import Dataset, DatasetAdapter, StatisticRole
from .abstract import CompareExtension


class StatTest(CompareExtension):
    def __init__(
        self, test_function: Callable | None = None, reliability: float = 0.05
    ):
        super().__init__()
        self.test_function = test_function
        self.reliability = reliability

    @staticmethod  # TODO: remove
    def check_other(other: Dataset | None) -> Dataset:
        if other is None:
            raise ValueError("No other dataset provided")
        return other

    @staticmethod
    def check_dataset(data: Dataset):
        if len(data.columns) != 1:
            raise ValueError("Data must be one-dimensional")

    def check_data(self, data: Dataset, other: Dataset | None) -> Dataset:
        other = self.check_other(other)

        self.check_dataset(data)
        self.check_dataset(other)

        return other

    def _calc_pandas(
        self, data: Dataset, other: Dataset | None = None, **kwargs
    ) -> Dataset | float:
        other = self.check_data(data, other)
        if self.test_function is None:
            raise ValueError("test_function is needed for execution")
        one_result = self.test_function(
            data.backend.data.values.flatten(),
            other.backend.data.values.flatten(),
            **kwargs,
        )
        one_result = DatasetAdapter.to_dataset(
            {
                "p-value": one_result.pvalue,
                "statistic": one_result.statistic,
                "pass": one_result.pvalue < self.reliability,
            },
            StatisticRole(),
        )
        return one_result


class TTestExtension(StatTest):
    def __init__(self, reliability: float = 0.05):
        super().__init__(ttest_ind, reliability=reliability)

    def _calc_pandas(
        self, data: Dataset, other: Dataset | None = None, **kwargs
    ) -> Dataset | float:
        # if (
        #     next(iter(data.nunique().values()))
        #     and next(iter(other.nunique().values())) < 2
        # ):
        #     return DatasetAdapter.to_dataset(
        #         {
        #             "p-value": [None],
        #             "statistic": [None],
        #             "pass": [None],
        #         },
        #         StatisticRole(),
        #     )
        return super()._calc_pandas(data, other, nan_policy="omit", **kwargs)


class KSTestExtension(StatTest):
    def __init__(self, reliability: float = 0.05):
        super().__init__(ks_2samp, reliability=reliability)


class UTestExtension(StatTest):
    def __init__(self, reliability: float = 0.05):
        super().__init__(mannwhitneyu, reliability=reliability)


class Chi2TestExtension(StatTest):
    @staticmethod
    def mini_category_replace(counts: Dataset) -> Dataset:
        mini_counts = counts["count"][counts["count"] < 7]
        if len(mini_counts) > 0:
            counts = counts.append(
                Dataset.from_dict(
                    [{counts.columns[0]: "other", "count": mini_counts["count"].sum()}],
                    roles=mini_counts.roles,
                )
            )
            counts = counts[counts["count"] >= 7]
        return counts

    def matrix_preparation(self, data: Dataset, other: Dataset) -> Dataset | None:
        proportion = len(data) / (len(data) + len(other))
        counted_data = data.value_counts()
        counted_data = self.mini_category_replace(counted_data)
        data_vc = counted_data["count"] * (1 - proportion)

        counted_other = other.value_counts()
        counted_other = self.mini_category_replace(counted_other)
        other_vc = counted_other["count"] * proportion

        if len(counted_data) < 2:
            return None
        data_vc = data_vc.add_column(counted_data[counted_data.columns[0]])
        other_vc = other_vc.add_column(counted_data[counted_data.columns[0]])
        return data_vc.merge(other_vc, on=counted_data.columns[0])[
            ["count_x", "count_y"]
        ].fillna(0)

    def _calc_pandas(
        self, data: Dataset, other: Dataset | None = None, **kwargs
    ) -> Dataset | float:
        other = self.check_data(data, other)
        matrix = self.matrix_preparation(data, other)
        if matrix is None:
            warnings.warn(f"Matrix Chi2 is empty for {data.columns[0]}. Returning None")
            return DatasetAdapter.to_dataset(
                {
                    "p-value": [None],
                    "statistic": [None],
                    "pass": [None],
                },
                StatisticRole(),
            )
        one_result = chi2_contingency(matrix.backend.data)
        return DatasetAdapter.to_dataset(
            {
                "p-value": (
                    one_result[1]
                    if isinstance(one_result, tuple)
                    else one_result.pvalue
                ),
                "statistic": (
                    one_result[0]
                    if isinstance(one_result, tuple)
                    else one_result.statistic
                ),
                "pass": (
                    one_result[1]
                    if isinstance(one_result, tuple)
                    else one_result.pvalue
                )
                < self.reliability,
            },
            StatisticRole(),
        )


class NormCDF(StatTest):
    def _calc_pandas(
        self, data: Dataset, other: Dataset | None = None, **kwargs
    ) -> Dataset | float:
        result = norm.cdf(abs(data.get_values()[0][0]))
        return DatasetAdapter.to_dataset(
            {"p-value": 2 * (1 - result)},
            StatisticRole(),
        )



from __future__ import annotations

from ..dataset import Dataset
from ..extensions.encoders import DummyEncoderExtension
from .abstract import Encoder


class DummyEncoder(Encoder):
    @staticmethod
    def _inner_function(
        data: Dataset, target_cols: str | None = None, **kwargs
    ) -> Dataset:
        if not target_cols:
            return Dataset.create_empty()
        return DummyEncoderExtension().calc(
            data=data, target_cols=target_cols, **kwargs
        )



from __future__ import annotations

import contextlib
from typing import Any, ClassVar

from ..comparators import Chi2Test, GroupDifference, GroupSizes, KSTest, TTest
from ..dataset import Dataset, ExperimentData, InfoRole, StatisticRole
from ..splitters import AASplitter, AASplitterWithStratification
from ..utils import ID_SPLIT_SYMBOL, ExperimentDataEnum, NotFoundInExperimentDataError
from .abstract import Reporter, TestDictReporter


class OneAADictReporter(TestDictReporter):
    tests: ClassVar[list] = [TTest, KSTest, Chi2Test]

    @staticmethod
    def convert_flat_dataset(data: dict) -> Dataset:
        struct_dict = OneAADictReporter._get_struct_dict(data)
        return OneAADictReporter._convert_struct_dict_to_dataset(struct_dict)

    @staticmethod
    def get_splitter_id(data: ExperimentData):
        for c in [AASplitter, AASplitterWithStratification]:
            with contextlib.suppress(NotFoundInExperimentDataError):
                return data.get_one_id(c, ExperimentDataEnum.additional_fields)

    def extract_group_difference(self, data: ExperimentData) -> dict[str, Any]:
        group_difference_ids = data.get_ids(GroupDifference)[GroupDifference.__name__][
            ExperimentDataEnum.analysis_tables.value
        ]
        return self._extract_from_comparators(data, group_difference_ids)

    def extract_group_sizes(self, data: ExperimentData) -> dict[str, Any]:
        group_sizes_id = data.get_one_id(GroupSizes, ExperimentDataEnum.analysis_tables)
        return self._extract_from_comparators(data, [group_sizes_id])

    def extract_analyzer_data(self, data: ExperimentData) -> dict[str, Any]:
        analyzer_id = data.get_one_id(
            "OneAAStatAnalyzer", ExperimentDataEnum.analysis_tables
        )
        return self.extract_from_one_row_dataset(data.analysis_tables[analyzer_id])

    def extract_data_from_analysis_tables(self, data: ExperimentData) -> dict[str, Any]:
        result = {}
        result.update(self.extract_group_difference(data))
        # result.update(self.extract_group_sizes(data))
        result.update(self.extract_tests(data))
        result.update(self.extract_analyzer_data(data))
        if self.front:
            result = self.rename_passed(result)
        return result

    def report(self, data: ExperimentData) -> dict[str, Any]:
        result = {
            "splitter_id": self.get_splitter_id(data),
        }
        result.update(self.extract_data_from_analysis_tables(data))
        return result


class AADatasetReporter(OneAADictReporter):
    def report(self, data: ExperimentData):
        front_buffer = self.front
        self.front = False
        dict_report = super().report(data)
        self.front = front_buffer
        return self.convert_flat_dataset(dict_report)


class AAPassedReporter(Reporter):
    @staticmethod
    def _reformat_aa_score_table(table: Dataset) -> Dataset:
        result = {}
        for ind in table.index:
            splitted_index = ind.split(ID_SPLIT_SYMBOL)
            row_index = f"{splitted_index[0]}{ID_SPLIT_SYMBOL}{splitted_index[-1]}"
            value = table.get_values(ind, "pass")
            if row_index not in result:
                result[row_index] = {splitted_index[1]: value}
            else:
                result[row_index][splitted_index[1]] = value
        result = Dataset.from_dict(result, roles={}).transpose() * 1
        return result

    @staticmethod
    def _reformat_best_split_table(table: Dataset) -> Dataset:
        passed = table.loc[:, [c for c in table.columns if (c.endswith("pass"))]]
        new_index = table.apply(
            lambda x: f"{x['feature']}{ID_SPLIT_SYMBOL}{x['group']}",
            {"index": InfoRole()},
            axis=1,
        )
        passed.index = new_index.get_values(column="index")
        passed = passed.rename(
            names={c: c[: c.rfind("pass") - 1] for c in passed.columns}
        )
        passed = passed.replace("OK", 1).replace("NOT OK", 0)
        passed = passed.astype({c: int for c in passed.columns}, errors="ignore")
        return passed

    def _detect_pass(self, analyzer_tables: dict[str, Dataset]):
        score_table = self._reformat_aa_score_table(analyzer_tables["aa score"])
        best_split_table = self._reformat_best_split_table(
            analyzer_tables["best split statistics"]
        )
        resume_table = score_table * best_split_table
        resume_table = resume_table.apply(
            lambda x: "OK" if x.sum() > 0 else "NOT OK",
            axis=1,
            role={"result": StatisticRole()},
        )
        result = score_table.merge(
            best_split_table,
            suffixes=(" aa test", " best split"),
            left_index=True,
            right_index=True,
        )
        result = result.merge(resume_table, left_index=True, right_index=True)
        result.roles = {c: r.__class__(str) for c, r in result.roles.items()}
        result = (
            result.replace(0, "NOT OK")
            .replace(1, "OK")
            .replace("0", "NOT OK")
            .replace("1", "OK")
        )
        splitted_index = [str(i).split(ID_SPLIT_SYMBOL) for i in result.index]
        result.add_column([i[0] for i in splitted_index], role={"feature": InfoRole()})
        result.add_column([i[1] for i in splitted_index], role={"group": InfoRole()})
        result.index = range(len(splitted_index))
        return result

    def report(self, data: ExperimentData) -> Dataset:
        analyser_ids = data.get_ids(
            "AAScoreAnalyzer", ExperimentDataEnum.analysis_tables
        )
        analyser_tables = {
            id_[id_.rfind(ID_SPLIT_SYMBOL) + 1 :]: data.analysis_tables[id_]
            for id_ in analyser_ids["AAScoreAnalyzer"][
                ExperimentDataEnum.analysis_tables.value
            ]
        }
        if not analyser_tables["aa score"]:
            print("AA test cannot be performed as none of the analyzers passed")
            return None
        result = self._detect_pass(analyser_tables)
        stats_cols = [
            "feature",
            "group",
            "control mean",
            "test mean",
            "difference",
            "difference %",
        ]
        differences = analyser_tables["best split statistics"].loc[
            :,
            [
                col
                for col in stats_cols
                if col in analyser_tables["best split statistics"].columns
            ],
        ]
        result = result.merge(differences, on=["feature", "group"], how="left")
        result = result[
            ["feature", "group"]
            + [c for c in result.columns if c not in ["feature", "group"]]
        ]
        numeric_cols = ["control mean", "test mean", "difference", "difference %"]
        for col in numeric_cols:
            result.data[col] = result.data[col].astype(float).round(6)
        return result


class AABestSplitReporter(Reporter):
    def report(self, data: ExperimentData):
        best_split_id = next(
            (c for c in data.additional_fields.columns if c.endswith("best")), []
        )
        markers = data.additional_fields.loc[:, best_split_id]
        markers = markers.rename({markers.columns[0]: "split"})
        return data.ds.merge(markers, left_index=True, right_index=True)



from __future__ import annotations

from typing import Literal

from .analyzers.ab import ABAnalyzer
from .comparators import Chi2Test, GroupDifference, GroupSizes, KSTest, TTest, UTest
from .dataset import AdditionalTargetRole, TargetRole, TreatmentRole
from .executor.executor import Executor
from .experiments.base import Experiment, OnRoleExperiment
from .transformers import CUPEDTransformer
from .ui.ab import ABOutput
from .ui.base import ExperimentShell
from .utils import ABNTestMethodsEnum, ABTestTypesEnum


class ABTest(ExperimentShell):
    """A class for conducting A/B tests with configurable statistical tests and multiple testing correction.

    This class provides functionality to run A/B tests with options for different statistical tests
    (t-test, u-test, chi-square test) and multiple testing correction methods.

    Args:
        additional_tests (Union[str, ABTestTypesEnum, List[Union[str, ABTestTypesEnum]], None], optional): Statistical test(s) to run in addition to
            the default group difference calculation. Valid options are 't-test', 'u-test', 'chi2-test' or ABTestTypesEnum.t_test, ABTestTypesEnum.u_test, and ABTestTypesEnum.chi2_test.
            Can be a single test name/enum or list of test names/enums. Defaults to [ABTestTypesEnum.t_test].
        multitest_method (ABNTestMethodsEnum, optional): Method to use for multiple testing correction. Valid options are:
            ABNTestMethodsEnum.bonferroni, ABNTestMethodsEnum.sidak, etc. Defaults to ABNTestMethodsEnum.holm.

            For more information refer to the statsmodels documentation:
            https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html

    Examples
    --------
    .. code-block:: python

        # Basic A/B test with default t-test
        ab_test = ABTest()
        results = ab_test.execute(data)

        # A/B test with multiple statistical tests
        ab_test = ABTest(
            additional_tests=[ABTestTypesEnum.t_test, ABTestTypesEnum.chi2_test],
            multitest_method=ABNTestMethodsEnum.bonferroni,
            cuped_features={"target_feature": "pre_target_feature"},
            enable_cupac=True,
            cupac_models=['linear', 'ridge']
        )
        results = ab_test.execute(data)
    """

    @staticmethod
    def _make_experiment(
        additional_tests: str | ABTestTypesEnum | list[str | ABTestTypesEnum] | None,
        multitest_method: ABNTestMethodsEnum | str | None,
        cuped_features: dict[str, str] | None,
        cupac_models: str | list[str] | None,
        enable_cupac: bool,
    ) -> Experiment:
        test_mapping: dict[str, Executor] = {
            "t-test": TTest(compare_by="groups", grouping_role=TreatmentRole()),
            "ks-test": KSTest(compare_by="groups", grouping_role=TreatmentRole()),
            "u-test": UTest(compare_by="groups", grouping_role=TreatmentRole()),
            "chi2-test": Chi2Test(compare_by="groups", grouping_role=TreatmentRole()),
        }
        on_role_executors: list[Executor] = [
            GroupDifference(grouping_role=TreatmentRole())
        ]
        additional_tests = (
            [ABTestTypesEnum.t_test] if additional_tests is None else additional_tests
        )
        multitest_method = (
            ABNTestMethodsEnum(multitest_method)
            if (
                multitest_method is not None
                and multitest_method in ABNTestMethodsEnum.__members__.values()
            )
            else ABNTestMethodsEnum.holm
        )
        if additional_tests:
            if isinstance(additional_tests, list):
                additional_tests = [
                    ABTestTypesEnum(test) if isinstance(test, str) else test
                    for test in additional_tests
                ]
            else:
                additional_tests = (
                    ABTestTypesEnum(additional_tests)
                    if isinstance(additional_tests, str)
                    else additional_tests
                )
        additional_tests = (
            additional_tests
            if isinstance(additional_tests, list)
            else [additional_tests]
        )
        for test_name in additional_tests:
            on_role_executors.append(test_mapping[test_name.value])

        # Build base executors list
        executors: list[Executor] = [
            GroupSizes(grouping_role=TreatmentRole()),
            OnRoleExperiment(
                executors=on_role_executors,
                role=(
                    [TargetRole(), AdditionalTargetRole()]
                    if enable_cupac
                    else TargetRole()
                ),
            ),
            ABAnalyzer(
                multitest_method=(
                    ABNTestMethodsEnum(multitest_method) if multitest_method else None
                )
            ),
        ]
        if cuped_features:
            executors.insert(0, CUPEDTransformer(cuped_features=cuped_features))

        if enable_cupac:
            from .ml import CUPACExecutor

            executors.insert(0, CUPACExecutor(cupac_models=cupac_models))

        return Experiment(executors=executors)

    def __init__(
        self,
        additional_tests: (
            str | ABTestTypesEnum | list[str | ABTestTypesEnum] | None
        ) = None,
        multitest_method: (
            Literal[
                "bonferroni",
                "sidak",
                "holm-sidak",
                "holm",
                "simes-hochberg",
                "hommel",
                "fdr_bh",
                "fdr_by",
                "fdr_tsbh",
                "fdr_tsbhy",
                "quantile",
            ]
            | None
        ) = "holm",
        t_test_equal_var: bool | None = None,
        cuped_features: dict[str, str] | None = None,
        cupac_models: str | list[str] | None = None,
        enable_cupac: bool = False,
    ):
        """
        Args:
            additional_tests: Statistical test(s) to run in addition to the default group difference calculation. Valid options are 't-test', 'u-test', 'chi2-test' or ABTestTypesEnum.t_test, ABTestTypesEnum.u_test, and ABTestTypesEnum.chi2_test. Can be a single test name/enum or list of test names/enums. Defaults to [ABTestTypesEnum.t_test].
            multitest_method: Method to use for multiple testing correction. Valid options are ABNTestMethodsEnum.bonferroni, ABNTestMethodsEnum.sidak, etc. Defaults to ABNTestMethodsEnum.holm.
            t_test_equal_var: Whether to use equal variance in t-test (optional).
            cuped_features: dict[str, str] — Dictionary {target_feature: pre_target_feature} for CUPED. Only dict is allowed.
            cupac_models: str | list[str] — model name (e.g. 'linear', 'ridge', 'lasso', 'catboost') or list of model names to try. If None, all available models will be tried and the best will be selected by variance reduction.
            enable_cupac: bool — Enable CUPAC variance reduction. CUPAC configuration is extracted from dataset.features_mapping.
        """
        super().__init__(
            experiment=self._make_experiment(
                additional_tests,
                multitest_method,
                cuped_features,
                cupac_models,
                enable_cupac,
            ),
            output=ABOutput(),
        )
        if t_test_equal_var is not None:
            self.experiment.set_params(
                {TTest: {"calc_kwargs": {"equal_var": t_test_equal_var}}}
            )



from __future__ import annotations

from ..analyzers.ab import ABAnalyzer
from ..comparators import GroupDifference, GroupSizes
from ..dataset import Dataset, ExperimentData, InfoRole, StatisticRole, TreatmentRole
from ..reporters.ab import ABDatasetReporter
from ..utils import ID_SPLIT_SYMBOL, ExperimentDataEnum
from .base import Output


class CupacOutput:
    """Container for CUPAC-specific outputs.

    Attributes:
        variance_reductions (Dataset | None): Variance reduction metrics from CUPAC models.
        feature_importances (Dataset | None): Feature importance scores from CUPAC models.
    """

    def __init__(self):
        self.variance_reductions: Dataset | None = None
        self.feature_importances: Dataset | None = None

    def __repr__(self) -> str:
        has_vr = self.variance_reductions is not None
        has_fi = self.feature_importances is not None

        if not has_vr and not has_fi:
            return "CupacOutput(no CUPAC data available)"

        parts = []
        if has_vr:
            n_targets = len(self.variance_reductions.data)
            parts.append(f"variance_reductions: {n_targets} target(s)")
        if has_fi:
            n_features = len(self.feature_importances.data)
            parts.append(f"feature_importances: {n_features} feature(s)")

        return f"CupacOutput({', '.join(parts)})"


class ABOutput(Output):
    multitest: Dataset | str
    sizes: Dataset
    cupac: CupacOutput

    def __init__(self):
        self._groups = []
        self.cupac = CupacOutput()
        super().__init__(resume_reporter=ABDatasetReporter())

    def _extract_multitest_result(self, experiment_data: ExperimentData):
        multitest_id = experiment_data.get_one_id(
            ABAnalyzer, ExperimentDataEnum.analysis_tables
        )
        if multitest_id and "MultiTest" in multitest_id:
            self.multitest = experiment_data.analysis_tables[multitest_id]
        else:
            self.multitest = (
                "There was less than three groups or multitest method wasn't provided"
            )

    def _extract_differences(self, experiment_data: ExperimentData):
        targets = []
        groups = []
        ids = experiment_data.get_ids(
            GroupDifference,
            searched_space=ExperimentDataEnum.analysis_tables,
        )["GroupDifference"]["analysis_tables"]
        self._groups = list(
            experiment_data.groups[
                experiment_data.ds.search_columns(TreatmentRole())[0]
            ].keys()
        )[1:]
        for i in self._groups:
            groups += [i] * len(ids)
        diff = Dataset.create_empty()
        for i in range(len(ids)):
            diff = diff.append(experiment_data.analysis_tables[ids[i]])
            targets += [ids[i].split(ID_SPLIT_SYMBOL)[-1]]
        return diff.add_column(groups, role={"group": StatisticRole()}).add_column(
            targets * len(self._groups), role={"feature": StatisticRole()}
        )

    def _extract_sizes(self, experiment_data: ExperimentData):
        ids = experiment_data.get_ids(
            GroupSizes,
            searched_space=ExperimentDataEnum.analysis_tables,
        )["GroupSizes"]["analysis_tables"]
        self.sizes = experiment_data.analysis_tables[ids[0]].add_column(
            self._groups, role={"group": StatisticRole()}
        )

    def _extract_variance_reductions(self, experiment_data: ExperimentData):
        """Extract variance reduction data from analysis_tables."""
        # Find all CUPAC report keys in analysis_tables
        cupac_report_keys = [
            key
            for key in experiment_data.analysis_tables.keys()
            if key.endswith("_cupac_report")
        ]

        if not cupac_report_keys:
            self.cupac.variance_reductions = None
            return

        # Aggregate all CUPAC reports into a single dataset
        variance_data = []
        for key in cupac_report_keys:
            report = experiment_data.analysis_tables[key]
            target_name = key.replace("_cupac_report", "")

            control_mean_bias = None
            test_mean_bias = None

            resume_data = self.resume.data
            if (
                "feature" in resume_data.columns
                and target_name in resume_data["feature"].values
            ):
                original_row = resume_data[resume_data["feature"] == target_name]
                cupac_row = resume_data[
                    resume_data["feature"] == f"{target_name}_cupac"
                ]

                control_mean_bias = (
                    original_row["control mean"].iloc[0]
                    - cupac_row["control mean"].iloc[0]
                )
                test_mean_bias = (
                    original_row["test mean"].iloc[0] - cupac_row["test mean"].iloc[0]
                )

            variance_data.append(
                {
                    "target": target_name,
                    "best_model": report.get("cupac_best_model"),
                    "variance_reduction_cv": report.get("cupac_variance_reduction_cv"),
                    "variance_reduction_real": report.get(
                        "cupac_variance_reduction_real"
                    ),
                    "control_mean_bias": control_mean_bias,
                    "test_mean_bias": test_mean_bias,
                }
            )

        self.cupac.variance_reductions = Dataset.from_dict(
            data=variance_data,
            roles={
                "target": InfoRole(str),
                "best_model": InfoRole(str),
                "variance_reduction_cv": StatisticRole(),
                "variance_reduction_real": StatisticRole(),
                "control_mean_bias": StatisticRole(),
                "test_mean_bias": StatisticRole(),
            },
        )

    def _extract_feature_importances(self, experiment_data: ExperimentData):
        """Extract feature importances from CUPAC models."""
        # Find all CUPAC report keys in analysis_tables
        cupac_report_keys = [
            key
            for key in experiment_data.analysis_tables.keys()
            if key.endswith("_cupac_report")
        ]

        if not cupac_report_keys:
            self.cupac.feature_importances = None
            return

        # Aggregate all feature importances into a single dataset
        importance_data = []
        for key in cupac_report_keys:
            report = experiment_data.analysis_tables[key]
            target_name = key.replace("_cupac_report", "")
            model_name = report.get("cupac_best_model")
            importances = report.get("cupac_feature_importances", {})

            if not importances:
                continue

            # Convert feature importances to rows
            for feature_idx, importance_value in importances.items():
                importance_data.append(
                    {
                        "target": target_name,
                        "feature": feature_idx,
                        "importance": importance_value,
                        "model": model_name,
                    }
                )

        if not importance_data:
            self.cupac.feature_importances = None
            return

        self.cupac.feature_importances = Dataset.from_dict(
            data=importance_data,
            roles={
                "target": InfoRole(str),
                "feature": InfoRole(str),
                "importance": StatisticRole(),
                "model": InfoRole(str),
            },
        )

    @property
    def variance_reduction_report(self) -> Dataset | str:
        """Get variance reduction report for CUPED/CUPAC transformations."""
        if hasattr(self, "_experiment_data"):
            return self.resume_reporter.report_variance_reductions(
                self._experiment_data
            )
        return "No experiment data available."

    def extract(self, experiment_data: ExperimentData):
        super().extract(experiment_data)
        self._extract_differences(experiment_data)
        self._extract_multitest_result(experiment_data)
        self._extract_sizes(experiment_data)
        self._extract_variance_reductions(experiment_data)
        self._extract_feature_importances(experiment_data)



from .encoders.encoders import DummyEncoder
from .experiments.base import Experiment
from .transformers.category_agg import CategoryAggregator
from .transformers.filters import (
    ConstFilter,
    CorrFilter,
    CVFilter,
    NanFilter,
    OutliersFilter,
)
from .transformers.na_filler import NaFiller

PREPROCESSING_DATA = Experiment(
    executors=[
        NaFiller(method="ffill"),
        CategoryAggregator(),
        CorrFilter(),
        CVFilter(),
        NanFilter(),
        ConstFilter(),
        OutliersFilter(lower_percentile=0.05, upper_percentile=0.95),
        DummyEncoder(),
    ]
)



from __future__ import annotations

from ..dataset import Dataset
from ..extensions.scipy_stats import (
    Chi2TestExtension,
    KSTestExtension,
    TTestExtension,
    UTestExtension,
)
from ..utils.constants import NUMBER_TYPES_LIST
from .abstract import StatHypothesisTesting


class TTest(StatHypothesisTesting):
    @property
    def search_types(self) -> list[type] | None:
        return NUMBER_TYPES_LIST

    @classmethod
    def _inner_function(
        cls, data: Dataset, test_data: Dataset | None = None, **kwargs
    ) -> Dataset:
        return TTestExtension(kwargs.get("reliability", 0.05)).calc(
            data, other=test_data, **kwargs
        )


class KSTest(StatHypothesisTesting):
    @property
    def search_types(self) -> list[type] | None:
        return NUMBER_TYPES_LIST

    @classmethod
    def _inner_function(
        cls, data: Dataset, test_data: Dataset | None = None, **kwargs
    ) -> Dataset:
        return KSTestExtension(kwargs.get("reliability", 0.05)).calc(
            data, other=test_data, **kwargs
        )


class UTest(StatHypothesisTesting):
    @property
    def search_types(self) -> list[type] | None:
        return NUMBER_TYPES_LIST

    @classmethod
    def _inner_function(
        cls, data: Dataset, test_data: Dataset | None = None, **kwargs
    ) -> Dataset:
        return UTestExtension(kwargs.get("reliability", 0.05)).calc(
            data, other=test_data, **kwargs
        )


class Chi2Test(StatHypothesisTesting):
    @property
    def search_types(self) -> list[type] | None:
        return [str]

    @classmethod
    def _inner_function(
        cls, data: Dataset, test_data: Dataset | None = None, **kwargs
    ) -> Dataset:
        return Chi2TestExtension(reliability=kwargs.get("reliability", 0.05)).calc(
            data, other=test_data, **kwargs
        )



# import sys
#
# from ..analyzers import ABAnalyzer, OneAAStatAnalyzer
# from ..comparators import GroupDifference, GroupSizes, ATE, TTest, KSTest, UTest
# from ..dataset import (
#     ExperimentData,
#     Arg1Role,
#     Arg2Role,
#     InfoRole,
#     TargetRole,
#     FeatureRole,
#     GroupingRole,
#     PreTargetRole,
#     StatisticRole,
#     StratificationRole,
#     TreatmentRole,
#     TempTreatmentRole,
#     TempGroupingRole,
#     TempTargetRole,
# )
# from ..experiments import (
#     Experiment,
#     OnRoleExperiment,
#     GroupExperiment,
#     CycledExperiment,
# )
# from ..reporters import OneAADictReporter
# from ..transformers import Shuffle
# from ..utils import ExperimentDataEnum, SpaceEnum
#
# all_classes = [
#     ABAnalyzer,
#     OneAAStatAnalyzer,
#     GroupDifference,
#     GroupSizes,
#     ATE,
#     TTest,
#     KSTest,
#     UTest,
#     Arg1Role,
#     Arg2Role,
#     InfoRole,
#     TargetRole,
#     FeatureRole,
#     GroupingRole,
#     PreTargetRole,
#     StatisticRole,
#     StratificationRole,
#     TreatmentRole,
#     TempTreatmentRole,
#     TempGroupingRole,
#     TempTargetRole,
#     OnRoleExperiment,
#     GroupExperiment,
#     CycledExperiment,
#     OneAADictReporter,
#     Shuffle,
#     ExperimentDataEnum,
#     SpaceEnum,
# ]
#
# spaces = {
#     "additional": SpaceEnum.additional,
#     "auto": SpaceEnum.auto,
#     "data": SpaceEnum.data,
# }
#
#
# class Factory:
#     def __init__(self, hypothesis):
#         self.hypothesis = hypothesis
#
#     def make_experiment(self, experiment):
#         executors = []
#         for key, items in experiment.items():
#             class_ = getattr(sys.modules[__name__], key)
#             if "executors" in items or "inner_executors" in items:
#                 item = "executors" if "executors" in items else "inner_executors"
#                 items[f"{item}"] = self.make_experiment(experiment[key][f"{item}"][0])
#             if "role" in items or "grouping_role" in items:
#                 item = "role" if "role" in items else "grouping_role"
#                 items[f"{item}"] = getattr(
#                     sys.modules[__name__], items[item] + "Role"
#                 )()
#             if "space" in items:
#                 items["space"] = spaces.get(items["space"])
#             items = {i: None if j == "None" else j for i, j in items.items()}
#             executors.append(class_(**items))
#         return executors
#
#     def execute(self):
#         experiment_data = ExperimentData(self.hypothesis.dataset)
#         experiment = Experiment(
#             executors=self.make_experiment(self.hypothesis.experiment)
#         )
#         return experiment_data, experiment



from .abstract import DatasetReporter, DictReporter, Reporter
from .homo import HomoDatasetReporter, HomoDictReporter

__all__ = [
    "DatasetReporter",
    "DictReporter",
    "HomoDatasetReporter",
    "HomoDictReporter",
    "Reporter",
]



import numpy as np
import pandas as pd  # type: ignore

from ..dataset import Dataset
from ..dataset.roles import FeatureRole
from .abstract import Extension


class CholeskyExtension(Extension):
    def _calc_pandas(self, data: Dataset, epsilon: float = 1e-3, **kwargs):
        cov = data.data.to_numpy()
        cov = cov + np.eye(cov.shape[0]) * epsilon
        return self.result_to_dataset(
            pd.DataFrame(np.linalg.cholesky(cov), columns=data.columns),
            {column: FeatureRole() for column in data.columns},
        )


class InverseExtension(Extension):
    def _calc_pandas(self, data: Dataset, **kwargs):
        return self.result_to_dataset(
            pd.DataFrame(np.linalg.inv(data.data.to_numpy()), columns=data.columns),
            {column: FeatureRole() for column in data.columns},
        )



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Sequence

from ..dataset import (
    ABCRole,
    AdditionalTargetRole,
    Dataset,
    ExperimentData,
    GroupingRole,
    TargetRole,
)
from ..executor import Calculator
from ..utils import AbstractMethodError, ExperimentDataEnum, NotSuitableFieldError
from ..utils.adapter import Adapter


class GroupOperator(
    Calculator
):  # TODO: change the derive from Calculator to COmparator
    def __init__(
        self,
        grouping_role: ABCRole | None = None,
        target_roles: ABCRole | list[ABCRole] | None = None,
        key: Any = "",
    ):
        super().__init__(key=key)
        self.target_roles = target_roles or TargetRole()
        self.grouping_role = grouping_role or GroupingRole()

    @property
    def search_types(self):
        return None

    @classmethod
    @abstractmethod
    def _inner_function(
        cls, data: Dataset, test_data: Dataset | None = None, **kwargs
    ) -> Any:
        raise AbstractMethodError

    def _get_fields(self, data: ExperimentData):
        group_field = data.field_search(self.grouping_role)
        target_fields = data.field_search(
            self.target_roles, search_types=self.search_types
        )
        if len(target_fields) != 2:
            target_fields += data.field_search(
                AdditionalTargetRole(), search_types=self.search_types
            )
        return group_field, target_fields

    @classmethod
    def _execute_inner_function(
        cls,
        grouping_data,
        target_fields: list[str] | None = None,
        **kwargs,
    ) -> dict:
        if target_fields is None or len(target_fields) != 2:
            raise ValueError(
                f"This operator works with 2 targets, but got {len(target_fields) if target_fields else None}"
            )
        result = {}
        for group, group_data in grouping_data:
            result[group[0]] = cls._inner_function(
                data=group_data[target_fields[0]],
                test_data=group_data[target_fields[1]],
                **kwargs,
            )
        return result

    @classmethod
    def calc(
        cls,
        data: Dataset,
        group_field: Sequence[str] | str | None = None,
        grouping_data: list[tuple[str, Dataset]] | None = None,
        target_fields: str | list[str] | None = None,
        **kwargs,
    ) -> dict:
        group_field = Adapter.to_list(group_field)

        if grouping_data is None:
            grouping_data = data.groupby(group_field)
        if len(grouping_data) > 1:
            grouping_data[0][1].tmp_roles = data.tmp_roles
        else:
            raise NotSuitableFieldError(group_field, "Grouping")
        return cls._execute_inner_function(
            grouping_data, target_fields=target_fields, old_data=data, **kwargs
        )

    def _set_value(
        self, data: ExperimentData, value: dict | None = None, key: Any = None
    ) -> ExperimentData:
        data.set_value(
            ExperimentDataEnum.variables,
            self.id,
            value,
        )
        return data



from .abstract import Comparator, StatHypothesisTesting
from .comparators import PSI, GroupDifference, GroupSizes
from .distances import MahalanobisDistance
from .hypothesis_testing import Chi2Test, KSTest, TTest, UTest
from .power_testing import MDEBySize, PowerTesting

__all__ = [
    "PSI",
    "Chi2Test",
    "Comparator",
    "GroupDifference",
    "GroupSizes",
    "KSTest",
    "MDEBySize",
    "MahalanobisDistance",
    "PowerTesting",
    "StatHypothesisTesting",
    "TTest",
    "UTest",
]



from __future__ import annotations

from abc import ABC
from copy import deepcopy

from ..utils import (
    CategoricalTypes,
    DefaultRoleTypes,
    FeatureRoleTypes,
    RoleNameType,
    TargetRoleTypes,
)


class ABCRole(ABC):
    _role_name: RoleNameType = "Abstract"

    def __init__(self, data_type: DefaultRoleTypes | None = None):
        self.data_type = data_type

    @property
    def role_name(self) -> str:
        return self._role_name

    def __repr__(self) -> str:
        return f"{self._role_name}({self.data_type})"

    def astype(self, data_type: DefaultRoleTypes | None = None) -> ABCRole:
        role = deepcopy(self)
        role.data_type = data_type
        return role

    def asadditional(self, data_type: DefaultRoleTypes | None = None) -> ABCRole:
        data_type = data_type or self.data_type
        for role_type in list(default_roles.values()):
            if isinstance(role_type, self.__class__) and isinstance(
                role_type, AdditionalRole
            ):
                return role_type.__class__(data_type)
        return self.__class__(data_type)


class LagRole(ABCRole):
    """Base class for roles that support temporal metadata (parent, lag)."""

    def __init__(
        self,
        data_type: DefaultRoleTypes | None = None,
        parent: str | None = None,
        lag: int | None = None,
    ):
        super().__init__(data_type)
        self.parent = parent
        self.lag = lag

    def __repr__(self) -> str:
        parts = []
        if self.data_type is not None:
            parts.append(f"data_type={self.data_type}")
        if self.parent is not None:
            parts.append(f"parent='{self.parent}'")
        if self.lag is not None:
            parts.append(f"lag={self.lag}")
        return (
            f"{self._role_name}({', '.join(parts)})"
            if parts
            else f"{self._role_name}()"
        )


class InfoRole(ABCRole):
    _role_name: RoleNameType = "Info"


class StratificationRole(ABCRole):
    _role_name: RoleNameType = "Stratification"

    def __init__(self, data_type: CategoricalTypes | None = None):
        super().__init__(data_type)


class GroupingRole(ABCRole):
    _role_name: RoleNameType = "Grouping"

    def __init__(self, data_type: CategoricalTypes | None = None):
        super().__init__(data_type)


class TreatmentRole(ABCRole):
    _role_name: RoleNameType = "Treatment"


class TargetRole(ABCRole):
    _role_name: RoleNameType = "Target"

    def __init__(
        self,
        data_type: TargetRoleTypes | None = None,
        cofounders: list[str] | None = None,
    ):
        super().__init__(data_type=data_type)
        self.cofounders = cofounders if cofounders is not None else []


class FeatureRole(LagRole):
    _role_name: RoleNameType = "Feature"

    def __init__(
        self,
        data_type: FeatureRoleTypes | None = None,
        parent: str | None = None,
        lag: int | None = None,
    ):
        super().__init__(data_type=data_type, parent=parent, lag=lag)


class PreTargetRole(LagRole):
    _role_name: RoleNameType = "PreTarget"

    def __init__(
        self,
        data_type: TargetRoleTypes | None = None,
        parent: str | None = None,
        lag: int | None = None,
        cofounders: list[str] | None = None,
    ):
        super().__init__(data_type=data_type, parent=parent, lag=lag)
        self.cofounders = cofounders if cofounders is not None else []


class StatisticRole(ABCRole):
    _role_name: RoleNameType = "Statistic"


class ResumeRole(ABCRole):
    _role_name = "Resume"


class FilterRole(ABCRole):
    _role_name: RoleNameType = "Filter"


class ConstGroupRole(ABCRole):
    _role_name: RoleNameType = "ConstGroup"


# ___________________________________________________________________________________________
class TempRole(ABCRole):
    _role_name: RoleNameType = "Temp"


class TempTreatmentRole(TempRole, TreatmentRole):
    _role_name: RoleNameType = "TempTreatment"


class TempTargetRole(TempRole, TargetRole):
    _role_name: RoleNameType = "TempTarget"


class TempGroupingRole(TempRole, GroupingRole):
    _role_name: RoleNameType = "TempGrouping"


class DefaultRole(ABCRole):
    _role_name: RoleNameType = "Default"


class ReportRole(ABCRole):
    _role_name: RoleNameType = "Report"


# ___________________________________________________________________________________________
class AdditionalRole(ABCRole):
    _role_name: RoleNameType = "Additional"


class AdditionalTreatmentRole(AdditionalRole, TreatmentRole):
    _role_name: RoleNameType = "AdditionalTreatment"


class AdditionalGroupingRole(AdditionalRole, GroupingRole):
    _role_name: RoleNameType = "AdditionalGrouping"


class AdditionalTargetRole(AdditionalRole, TargetRole):
    _role_name: RoleNameType = "AdditionalTarget"


class AdditionalFeatureRole(AdditionalRole, FeatureRole):
    _role_name: RoleNameType = "AdditionalTarget"


class AdditionalPreTargetRole(AdditionalRole, PreTargetRole):
    _role_name: RoleNameType = "AdditionalPreTarget"


class AdditionalMatchingRole(AdditionalRole):
    _role_name: RoleNameType = "AdditionalMatching"


default_roles: dict[RoleNameType, ABCRole] = {
    "info": InfoRole(),
    "default": DefaultRole(),
    "feature": FeatureRole(),
    "treatment": TreatmentRole(),
    "grouping": GroupingRole(),
    "target": TargetRole(),
    "pretarget": PreTargetRole(),
    "stratification": StratificationRole(),
    "statistic": StatisticRole(),
    "filter": FilterRole(),
    "constgroup": ConstGroupRole(),
    "additionaltreatment": AdditionalTreatmentRole(),
    "additionalgrouping": AdditionalGroupingRole(),
    "additionaltarget": AdditionalTargetRole(),
    "additionalfeature": AdditionalFeatureRole(),
    "additionalpretarget": AdditionalPreTargetRole(),
}



from __future__ import annotations

from copy import deepcopy
from typing import Any

from ..comparators import TTest, UTest
from ..dataset import Dataset, ExperimentData, StatisticRole, TargetRole, TreatmentRole
from ..experiments.base import Executor
from ..extensions.statsmodels import MultiTest, MultitestQuantile
from ..utils import (
    ID_SPLIT_SYMBOL,
    NAME_BORDER_SYMBOL,
    ABNTestMethodsEnum,
    BackendsEnum,
    ExperimentDataEnum,
)


class ABAnalyzer(Executor):
    def __init__(
        self,
        multitest_method: ABNTestMethodsEnum | None = None,
        alpha: float = 0.05,
        equal_variance: bool = True,
        quantiles: float | list[float] | None = None,
        iteration_size: int = 20000,
        random_state: int | None = None,
        key: Any = "",
    ):
        self.multitest_method = multitest_method
        self.alpha = alpha
        self.equal_variance = equal_variance
        self.quantiles = quantiles
        self.iteration_size = iteration_size
        self.random_state = random_state
        super().__init__(key)

    def _set_value(self, data: ExperimentData, value, key=None) -> ExperimentData:
        return data.set_value(
            ExperimentDataEnum.analysis_tables,
            self.id + key if key else self.id,
            value,
        )

    def execute_multitest(self, data: ExperimentData, p_values: Dataset, **kwargs):
        group_field = data.ds.search_columns(TreatmentRole())[0]
        target_fields = data.ds.search_columns(TargetRole(), search_types=[int, float])
        if self.multitest_method and len(data.groups[group_field]) > 2:
            if self.multitest_method != ABNTestMethodsEnum.quantile:
                multitest_result = MultiTest(self.multitest_method).calc(
                    p_values, **kwargs
                )
                groups = []
                for i in list(data.groups[group_field].keys())[1:]:
                    groups += [i] * len(target_fields)
                multitest_result = multitest_result.add_column(
                    groups
                    * (
                        len(multitest_result)
                        // len(target_fields)
                        // (len(data.groups[group_field]) - 1)
                    ),
                    role={"group": StatisticRole()},
                )

            else:
                multitest_result = Dataset.create_empty()
                for target_field in target_fields:
                    multitest_result = multitest_result.append(
                        MultitestQuantile(
                            self.alpha,
                            self.iteration_size,
                            self.equal_variance,
                            self.random_state,
                        ).calc(
                            p_values,
                            group_field=group_field,
                            target_field=target_field,
                            quantiles=self.quantiles,
                        )
                    )
            return self._set_value(data, multitest_result, key="MultiTest")
        return data

    def _add_pvalues(self, multitest_pvalues, value, field):
        if (
            self.multitest_method
            and field == "p-value"
            and self.multitest_method != "quantile"
        ):
            multitest_pvalues = multitest_pvalues.append(value)
        return multitest_pvalues

    def execute(self, data: ExperimentData) -> ExperimentData:
        executor_ids = data.get_ids([TTest, UTest])
        num_groups = len(data.groups[data.ds.search_columns(TreatmentRole())[0]]) - 1
        groups = list(data.groups[data.ds.search_columns(TreatmentRole())[0]].items())
        multitest_pvalues = Dataset.create_empty()
        analysis_data = {}
        for c, spaces in executor_ids.items():
            analysis_ids = spaces.get("analysis_tables", [])
            if len(analysis_ids) == 0:
                continue
            t_data = deepcopy(data.analysis_tables[analysis_ids[0]])
            for aid in analysis_ids[1:]:
                t_data = t_data.append(data.analysis_tables[aid])
            if len(analysis_ids) < len(t_data):
                analysis_ids *= num_groups
            t_data.data.index = analysis_ids
            for f in ["p-value", "pass"]:
                for i in range(0, len(analysis_ids), len(analysis_ids) // num_groups):
                    value = t_data.iloc[i : i + len(analysis_ids) // num_groups][f]
                    multitest_pvalues = self._add_pvalues(multitest_pvalues, value, f)
                    analysis_data[f"{c} {f} {groups[i // num_groups + 1][0]}"] = (
                        value.mean()
                    )
            if c not in ["UTest", "TTest"]:
                indexes = t_data.index
                values = t_data.data.values.tolist()
                for idx, value in zip(indexes, values):
                    name = idx.split(ID_SPLIT_SYMBOL)[-1]
                    analysis_data[
                        f"{c} {name[name.find(NAME_BORDER_SYMBOL) + 1 : name.rfind(NAME_BORDER_SYMBOL)]}"
                    ] = value[0]

        analysis_dataset = Dataset.from_dict(
            [analysis_data],
            {f: StatisticRole() for f in analysis_data},
            BackendsEnum.pandas,
        )
        data = self.execute_multitest(
            data,
            (
                multitest_pvalues
                if not multitest_pvalues.is_empty()
                and self.multitest_method != ABNTestMethodsEnum.quantile
                else data.ds
            ),
        )

        return self._set_value(data, analysis_dataset)



from __future__ import annotations

from typing import Any

from ..dataset import Dataset, ExperimentData
from ..executor.executor import Calculator


class Shuffle(Calculator):
    def __init__(
        self,
        random_state: int | None = None,
        key: Any = "",
    ):
        super().__init__(key)
        self.random_state = random_state

    @staticmethod
    def _inner_function(data: Dataset, random_state: int | None = None) -> Dataset:
        return data.shuffle(random_state=random_state)

    def generate_params_hash(self):
        return f"{self.random_state}"

    def execute(self, data: ExperimentData) -> ExperimentData:
        result = data.copy(data=self.calc(data=data.ds, random_state=self.random_state))
        return result



from .__version__ import __version__
from .aa import AATest
from .ab import ABTest
from .homogeneity import HomogeneityTest
from .matching import Matching

__all__ = ["AATest", "ABTest", "HomogeneityTest", "Matching", "__version__"]



from __future__ import annotations

import copy

import pandas as pd  # type: ignore

from ..dataset import Dataset, DatasetAdapter
from .abstract import Extension


class DummyEncoderExtension(
    Extension
):  # TODO: role types are being rewritten, needs to be fixed
    @staticmethod
    def _calc_pandas(data: Dataset, target_cols: str | None = None, **kwargs):
        dummies_df = pd.get_dummies(
            data=data[target_cols].data, drop_first=True
        ).astype(int)
        # Setting roles to the dummies in additional fields based on the original
        # roles by searching based on the part of the dummy column name
        roles = {
            col: data.roles[col[: col.rfind("_")]].asadditional(int)
            for col in dummies_df.columns
        }
        new_roles = copy.deepcopy(roles)
        for role in roles.values():
            role.data_type = bool
        return DatasetAdapter.to_dataset(dummies_df, roles=new_roles)



from __future__ import annotations

import copy
import json  # type: ignore
from abc import ABC
from typing import Any, Iterable

import pandas as pd  # type: ignore

from ..utils import BackendsEnum, RoleColumnError
from .backends import PandasDataset
from .roles import ABCRole, DefaultRole, default_roles


def parse_roles(roles: dict) -> dict[str | int] | ABCRole:
    new_roles = {}
    roles = roles or {}
    for role in roles:
        r = default_roles.get(role, role)
        if isinstance(roles[role], list):
            for i in roles[role]:
                new_roles[i] = copy.deepcopy(r)
        else:
            new_roles[roles[role]] = copy.deepcopy(r)
    return new_roles or roles


class DatasetBase(ABC):
    @staticmethod
    def _select_backend_from_data(data):
        return PandasDataset(data)

    @staticmethod
    def _select_backend_from_str(data, backend):
        if backend == BackendsEnum.pandas:
            return PandasDataset(data)
        if backend is None:
            return PandasDataset(data)
        raise TypeError("Backend must be an instance of BackendsEnum")

    def _set_all_roles(self, roles):
        keys = list(roles.keys())
        for column in self.columns:
            if column not in keys:
                roles[column] = copy.deepcopy(self.default_role) or DefaultRole()
        return roles

    def _set_empty_types(self, roles):
        for column, role in roles.items():
            if role.data_type is None:
                role.data_type = self._backend.get_column_type(column)
            self._backend = self._backend.update_column_type(column, role.data_type)

    def __init__(
        self,
        roles: dict[ABCRole, list[str] | str] | dict[str, ABCRole],
        data: pd.DataFrame | str | None = None,
        backend: BackendsEnum | None = None,
        default_role: ABCRole | None = None,
    ):
        self._backend = (
            self._select_backend_from_str(data, backend)
            if backend
            else self._select_backend_from_data(data)
        )
        self.default_role = default_role
        roles = (
            parse_roles(roles)
            if any(isinstance(role, ABCRole) for role in roles.keys())
            else roles
        )
        if any(not isinstance(role, ABCRole) for role in roles.values()):
            raise TypeError("Roles must be instances of ABCRole type")
        if data is not None and any(
            i not in self._backend.columns for i in list(roles.keys())
        ):
            raise RoleColumnError(list(roles.keys()), self._backend.columns)
        if data is not None:
            roles = self._set_all_roles(roles)
            self._set_empty_types(roles)
        self._roles: dict[str, ABCRole] = roles
        self._tmp_roles: (
            dict[ABCRole, list[str] | str] | dict[list[str] | str] | ABCRole
        ) = {}

    def __repr__(self):
        return self.data.__repr__()

    def _repr_html_(self):
        return self.data._repr_html_()

    def __len__(self):
        return self._backend.__len__()

    def search_columns(
        self,
        roles: ABCRole | Iterable[ABCRole],
        tmp_role=False,
        search_types: list | None = None,
    ) -> list[str]:
        roles = roles if isinstance(roles, Iterable) else [roles]
        roles_for_search = self._tmp_roles if tmp_role else self.roles
        return [
            str(column)
            for column, role in roles_for_search.items()
            if any(
                isinstance(r, role.__class__)
                and (not search_types or role.data_type in search_types)
                for r in roles
            )
        ]

    def search_columns_by_type(
        self,
        search_types: list | type,
    ) -> list[str]:
        search_types = (
            search_types if isinstance(search_types, Iterable) else [search_types]
        )
        return [
            str(column)
            for column, role in self.roles.items()
            if any(role.data_type == t for t in search_types)
        ]

    def replace_roles(
        self,
        new_roles_map: dict[ABCRole | str] | ABCRole,
        tmp_role: bool = False,
        auto_roles_types: bool = False,
    ):
        new_roles_map = parse_roles(
            {
                role: (
                    self.search_columns(column, tmp_role)
                    if isinstance(column, ABCRole)
                    else column
                )
                for column, role in new_roles_map.items()
            }
        )

        new_roles = {
            column: new_roles_map[column] if column in new_roles_map else role
            for column, role in self.roles.items()
        }

        if tmp_role:
            self._tmp_roles = new_roles
        else:
            self.roles = new_roles
            if auto_roles_types:
                self._set_empty_types(new_roles_map)

        return self

    @property
    def index(self):
        return self._backend.index

    @property
    def data(self):
        return self._backend.data

    @property
    def roles(self):
        return self._roles

    @roles.setter
    def roles(self, value):
        self._set_roles(new_roles_map=value, temp_role=False)

    @data.setter
    def data(self, value):
        self._backend.data = value

    @property
    def columns(self):
        return self._backend.columns

    @property
    def shape(self):
        return self._backend.shape

    @property
    def tmp_roles(self):
        return self._tmp_roles

    @tmp_roles.setter
    def tmp_roles(self, value):
        self._set_roles(new_roles_map=value, temp_role=True)
        self._set_empty_types(self._tmp_roles)

    def to_dict(self):
        return {
            "backend": self._backend.name,
            "roles": {
                "role_names": list(map(lambda x: x, list(self.roles.keys()))),
                "columns": list(self.roles.values()),
            },
            "data": self._backend.to_dict(),
        }

    def to_numpy(self):
        return self._backend.to_numpy()

    def to_records(self):
        return self._backend.to_records()

    def to_json(self, filename: str | None = None):
        if not filename:
            return json.dumps(self.to_dict())
        with open(filename, "w") as file:
            json.dump(self.to_dict(), file)

    @property
    def backend(self):
        return self._backend

    def get_values(
        self,
        row: str | None = None,
        column: str | None = None,
    ) -> Any:
        return self._backend.get_values(row=row, column=column)

    def iget_values(
        self,
        row: int | None = None,
        column: int | None = None,
    ) -> Any:
        return self._backend.iget_values(row=row, column=column)

    def _set_roles(
        self,
        new_roles_map: dict[ABCRole, list[str] | str] | dict[list[str] | str] | ABCRole,
        temp_role: bool = False,
    ):
        if not new_roles_map:
            if not temp_role:
                return self.roles
            else:
                self._tmp_roles = {}
                return self

        keys, values = list(new_roles_map.keys()), list(new_roles_map.values())
        roles, columns_sets = (
            (keys, values) if isinstance(keys[0], ABCRole) else (values, keys)
        )

        new_roles = {}
        for role, columns in zip(roles, columns_sets):
            if isinstance(columns, list):
                for column in columns:
                    new_roles[column] = copy.deepcopy(role)
            else:
                new_roles[columns] = copy.deepcopy(role)

        if temp_role:
            self._tmp_roles = new_roles
        else:
            self._roles = new_roles

        return self



from ..dataset.dataset import DatasetAdapter, ExperimentData
from ..dataset.roles import StatisticRole
from ..executor.executor import Executor
from ..operators.operators import MatchingMetrics
from ..utils.enums import ExperimentDataEnum


class MatchingAnalyzer(Executor):
    def _set_value(self, data: ExperimentData, value, key=None) -> ExperimentData:
        return data.set_value(
            ExperimentDataEnum.analysis_tables, self.id, value, key=key
        )

    def execute(self, data: ExperimentData):
        variables = data.variables[
            data.get_one_id(MatchingMetrics, space=ExperimentDataEnum.variables)
        ]
        columns = ["Effect Size", "Standard Error", "P-value", "CI Lower", "CI Upper"]
        return self._set_value(
            data,
            DatasetAdapter.to_dataset(
                variables,
                {field: StatisticRole() for field in list(variables.keys())},
            ).transpose(roles={column: StatisticRole() for column in columns}),
        )



import enum


@enum.unique
class ExperimentDataEnum(enum.Enum):
    variables = "variables"
    additional_fields = "additional_fields"
    analysis_tables = "analysis_tables"
    groups = "groups"
    ml = "ml"


@enum.unique
class BackendsEnum(enum.Enum):
    pandas = "pandas"


@enum.unique
class SpaceEnum(enum.Enum):
    auto = "auto"
    additional = "additional"
    data = "data"


@enum.unique
class ABNTestMethodsEnum(enum.Enum):
    bonferroni = "bonferroni"
    sidak = "sidak"
    holm_sidak = "holm-sidak"
    holm = "holm"
    simes_hochberg = "simes-hochberg"
    hommel = "hommel"
    fdr_bh = "fdr_bh"
    fdr_by = "fdr_by"
    fdr_tsbh = "fdr_tsbh"
    fdr_tsbky = "fdr_tsbky"
    quantile = "quantile"


@enum.unique
class ABTestTypesEnum(enum.Enum):
    t_test = "t-test"
    ks_test = "ks-test"
    u_test = "u-test"
    chi2_test = "chi2-test"


@enum.unique
class RenameEnum(enum.Enum):
    all = "all"
    columns = "columns"
    index = "index"



from __future__ import annotations

import sys
from pathlib import Path
from typing import Sequence

import numpy as np
import pandas as pd
from scipy import stats

ROOT = Path("").absolute().parents[0]
sys.path.append(str(ROOT))


class DataGenerator:
    """
    Advanced synthetic data generator with support for two lags for Y
    and control of correlation structure.
    """

    def __init__(
        self,
        n_samples=2000,
        distributions=None,
        time_correlations=None,
        effect_size=5.0,
        seed=None,
    ):
        self.n_samples = n_samples
        self.distributions = distributions or {
            "X1": {"type": "normal", "mean": 1, "std": 2},
            "X2": {"type": "bernoulli", "p": 0.4},
            "y0": {"type": "normal", "mean": 10, "std": 3},
        }
        self.time_correlations = time_correlations or {"X1": 0.7, "X2": 0.6, "y0": 0.8}
        self.effect_size = effect_size
        self.seed = seed
        np.random.seed(seed)

    def _generate_bernoulli_pair(self, p, rho):
        rho_max = min(p / (1 - p), (1 - p) / p)
        if abs(rho) > rho_max:
            raise ValueError(f"Impossible correlation {rho} for p={p}")
        p11 = p * p + rho * p * (1 - p)
        p10 = p * (1 - p) - rho * p * (1 - p)
        p01 = (1 - p) * p - rho * p * (1 - p)
        p00 = (1 - p) * (1 - p) + rho * p * (1 - p)
        states = np.random.choice(4, size=self.n_samples, p=[p00, p01, p10, p11])
        lag = (states == 1) | (states == 3)
        current = (states == 2) | (states == 3)
        return current.astype(int), lag.astype(int)

    def _generate_correlated_pair(self, dist_type, params, rho, U_vector=0):
        if dist_type == "normal":
            cov = [
                [params["std"] ** 2, rho * params["std"] ** 2],
                [rho * params["std"] ** 2, params["std"] ** 2],
            ]
            return (
                np.random.multivariate_normal(
                    [params["mean"], params["mean"]], cov, self.n_samples
                ).T
                + U_vector
            )
        elif dist_type == "bernoulli":
            return self._generate_bernoulli_pair(params["p"], rho)
        elif dist_type == "gamma":
            Z = np.random.multivariate_normal(
                [0, 0], [[1, rho], [rho, 1]], self.n_samples
            )
            U = stats.norm.cdf(Z)
            current = stats.gamma.ppf(U[:, 0], a=params["shape"], scale=params["scale"])
            lag = stats.gamma.ppf(U[:, 1], a=params["shape"], scale=params["scale"])
            return current, lag
        else:
            raise ValueError(f"Unsupported distribution: {dist_type}")

    def _generate_correlated_chain(self, params, rho, n_points, U=0):
        mean = params["mean"]
        std = params["std"]
        cov = np.zeros((n_points, n_points))
        for i in range(n_points):
            for j in range(n_points):
                cov[i, j] = (std**2) * (rho ** abs(i - j))
        return np.random.multivariate_normal(
            [mean] * n_points, cov, self.n_samples
        ).T  # + 0.1 * U

    def generate(self):
        data = {}
        data["z"] = np.random.binomial(1, 0.5, self.n_samples)
        data["U"] = np.random.normal(0, 1, self.n_samples)
        D_propensity = 0.3 + 0.4 * data["z"] + 0.3 * data["U"]
        data["D"] = np.random.binomial(1, np.clip(D_propensity, 0, 1))
        data["d"] = data["D"] * data["z"]
        for var in ["X1", "X2"]:
            current, lag = self._generate_correlated_pair(
                self.distributions[var]["type"],
                self.distributions[var],
                self.time_correlations[var],
                data["U"],
            )
            data[var] = current
            data[f"{var}_lag"] = lag
        y_params = self.distributions["y0"]
        y_rho = self.time_correlations["y0"]
        if y_params["type"] == "normal":
            y_chain = self._generate_correlated_chain(y_params, y_rho, 3, data["U"])
            data["y0"] = y_chain[2]
            data["y0_lag_1"] = y_chain[1]
            data["y0_lag_2"] = y_chain[0]
        else:
            current, lag1 = self._generate_correlated_pair(
                y_params["type"], y_params, y_rho
            )
            lag2, _ = self._generate_correlated_pair(y_params["type"], y_params, y_rho)
            data["y0"] = current
            data["y0_lag_1"] = lag1
            data["y0_lag_2"] = lag2
        data["y1"] = (
            data["y0"]
            + self.effect_size * (1 + data["U"])
            + np.random.normal(0, 0.01, self.n_samples)
        )
        data["y"] = np.where(data["d"] == 1, data["y1"], data["y0"])

        # Create DataFrame and rename columns for clearer temporal structure
        df = pd.DataFrame(data)
        df = df.rename(
            columns={
                "X1": "X1_lag1",  # X1 becomes period 1 covariate
                "X2": "X2_lag1",  # X2 becomes period 1 covariate
                "X1_lag": "X1_lag2",  # X1_lag becomes period 2 covariate
                "X2_lag": "X2_lag2",  # X2_lag becomes period 2 covariate
            }
        )
        return df


def set_nans(
    data: pd.DataFrame,
    na_step: Sequence[int] | int | None = None,
    nan_cols: Sequence[str] | str | None = None,
) -> pd.DataFrame:
    """Fills specific columns in a DataFrame with NaN values at given intervals.

    Args:
        data: The input DataFrame.
        na_step: Step interval(s) for inserting NaNs.
        nan_cols: Column(s) to insert NaNs.

    Returns:
        pd.DataFrame: Modified DataFrame with NaNs.
    """
    # Convert na_step and nan_cols to lists
    if na_step is None:
        na_step_list: list[int] = [10]
    elif isinstance(na_step, int):
        na_step_list = [na_step]
    else:
        na_step_list = list(na_step)

    if nan_cols is None:
        nan_cols_list: list[str] = list(data.columns)
    elif isinstance(nan_cols, str):
        nan_cols_list = [nan_cols]
    else:
        nan_cols_list = list(nan_cols)

    if not na_step_list or not nan_cols_list:
        raise ValueError("na_step and nan_cols must not be empty.")

    if len(na_step_list) < len(nan_cols_list):
        na_step_list += [na_step_list[-1]] * (len(nan_cols_list) - len(na_step_list))
    elif len(na_step_list) > len(nan_cols_list):
        na_step_list = na_step_list[: len(nan_cols_list)]

    for col, step in zip(nan_cols_list, na_step_list):
        if col in data.columns:
            data.loc[step::step, col] = None

    return data


def create_test_data(
    num_users: int = 10000,
    na_step: Sequence[int] | int | None = None,
    nan_cols: Sequence[str] | str | None = None,
    file_name: str | None = None,
    exact_ATT: int = 100,
    rs=None,
):
    """Creates data for tutorial.

    Args:
        num_users: num of strings
        na_step:
            num or list of nums of period to make NaN (step of range)
            If list - iterates accordingly order of columns
        nan_cols:
            name of one or several columns to fill with NaN
            If list - iterates accordingly order of na_step
        file_name: name of file to save; doesn't save file if None
        exact_ATT: an accurate synthetic effect

    Returns:
        data: dataframe with
    """
    if rs is not None:
        np.random.seed(rs)

    if (nan_cols is not None) and isinstance(nan_cols, str):
        nan_cols = [nan_cols]
    # Simulating dataset with known effect size
    num_months = 12

    # signup_months == 0 means customer did not sign up
    signup_months = np.random.choice(
        np.arange(1, num_months), num_users
    ) * np.random.randint(0, 2, size=num_users)

    data = pd.DataFrame(
        {
            "user_id": np.repeat(np.arange(num_users), num_months),
            "signup_month": np.repeat(
                signup_months, num_months
            ),  # signup month == 0 means customer did not sign up
            "month": np.tile(
                np.arange(1, num_months + 1), num_users
            ),  # months are from 1 to 12
            "spend": np.random.poisson(500, num_users * num_months),
        }
    )

    # A customer is in the treatment group if and only if they signed up
    data["treat"] = data["signup_month"] > 0

    # Simulating an effect of month (monotonically decreasing--customers buy less later in the year)
    data["spend"] = data["spend"] - data["month"] * 10

    # Simulating a simple treatment effect of 100
    after_signup = (data["signup_month"] < data["month"]) & (data["treat"])
    data.loc[after_signup, "spend"] = data[after_signup]["spend"] + exact_ATT

    # Setting the signup month (for ease of analysis)
    i = 3
    data = (
        data.groupby(["user_id", "signup_month", "treat"])
        .apply(
            lambda x: pd.Series(
                {
                    "pre_spends": x.loc[x.month < i, "spend"].mean(),
                    "post_spends": x.loc[x.month > i, "spend"].mean(),
                }
            )
        )
        .reset_index()
    )

    # Additional category features
    gender_i = np.random.choice(a=[0, 1], size=data.user_id.nunique())
    gender = [["M", "F"][i] for i in gender_i]

    age = np.random.choice(a=range(18, 70), size=data.user_id.nunique())

    industry_i = np.random.choice(a=range(1, 3), size=data.user_id.nunique())
    industry_names = ["Finance", "E-commerce", "Logistics"]
    industry = [industry_names[i] for i in industry_i]

    data["age"] = age
    data["gender"] = gender
    data["industry"] = industry
    data["industry"] = data["industry"].astype("str")
    data["treat"] = data["treat"].astype(int)

    # input nans in data if needed
    data = set_nans(data, na_step, nan_cols)

    if file_name is not None:
        data.to_csv(ROOT / f"{file_name}.csv", index=False)

    return data


def sigmoid(x: np.ndarray) -> np.ndarray:
    """Logistic sigmoid ufunc for ndarrays.

    The sigmoid function, also known as the logistic sigmoid function,
    is defined as sigmoid(x) = 1/(1+exp(-x)).
    It is the inverse of the logit function.

    Args:
        x: Input array.

    Returns:
        Sigmoid function of x.
    """
    return 1 / (1 + np.exp(-x))


def sigmoid_division(x, dependent_division=True) -> np.ndarray:
    """
    Division method via sigmoid function.

    Args:
        x:
            Input array
        dependent_division:
            If True - returns the binary vector dependent on the X.
            If False - returns the binary vector independent on the X.

    Returns:
        Binary array coordinated with variable x.
    """
    if dependent_division:
        division = np.random.binomial(1, sigmoid((x - x.mean()) / x.std()))
    else:
        division = np.random.binomial(1, 0.5, size=len(x))
    return division


def gen_special_medicine_df(
    data_size=100, *, dependent_division=True, random_state=None
) -> pd.DataFrame:
    """Synthetic dataframe generator.

    Realises dependent/independent group splitting.

    Args:
        data_size:
            Length of output Dataframe
        dependent_division:
            If True - the returned Dataframe contains a division into
            groups dependent on the features.
            If False - the returned Dataframe contains a division into
            groups independent on the features.
        random_state:
            If specified - defines numpy random seed. Defaults to None.

    Returns:
        Synthetic dataframe also containing fictional information about patient treatment.
    """
    if random_state is not None:
        np.random.seed(random_state)

    disease_degree = np.random.choice(
        [1, 2, 3, 4, 5], p=[0.3, 0.3, 0.2, 0.1, 0.1], size=data_size
    ).astype(int)

    age = np.random.normal(40, scale=8, size=data_size).astype(int)

    age_effect = (age ^ 2 - 400) / 1000

    experimental_treatment = sigmoid_division(disease_degree, dependent_division)

    residual_lifetime = np.random.exponential(
        17 - 2.5 * disease_degree + 1 * experimental_treatment - age_effect
    )

    df = pd.DataFrame(
        dict(
            age=age,
            disease_degree=disease_degree,
            experimental_treatment=experimental_treatment,
            residual_lifetime=residual_lifetime,
        )
    )
    return df


def gen_oracle_df(
    data_size=8, *, dependent_division=True, factual_only=False, random_state=None
) -> pd.DataFrame:
    """Synthetic dataframe generator.

    Realises factual and counterfactual outcomes.

    Args:
        data_size:
            Length of output Dataframe
        dependent_division:
            If True - the returned Dataframe contains a division into
            groups dependent on the features.
            If False - the returned Dataframe contains a division into
            groups independent on the features.
        factual_only:
            Defines the display of counterfactual states.
        random_state:
            If specified - defines numpy random seed. Defaults to None.

    Returns:
        Synthetic dataframe also containing fictional information about
        factual and counterfactual states of the target.
    """
    if random_state is not None:
        np.random.seed(random_state)

    treatment = np.random.binomial(1, 0.5, size=data_size)

    if dependent_division:
        target_feature = np.random.binomial(1, 0.3 + 0.4 * treatment)
    else:
        target_feature = np.random.binomial(1, 0.5, size=data_size)

    target_untreated = (
        np.random.uniform(low=300, high=800, size=data_size).round(-2).astype(int)
    )

    target_treated = target_untreated + 50 + target_feature * 100

    if factual_only:
        target_untreated = np.where(1 - treatment, target_untreated, np.nan)
        target_treated = np.where(treatment, target_treated, np.nan)

    y_factual = np.where(treatment, target_treated, target_untreated).astype("int")

    treatment_effect = target_treated - target_untreated

    df = pd.DataFrame(
        dict(
            X=target_feature,
            Target_untreated=target_untreated,
            Target_treated=target_treated,
            Treatment=treatment,
            Target=y_factual,
            TE=treatment_effect,
        )
    )
    return df


def gen_control_variates_df(
    data_size=1000, *, dependent_division=True, random_state=None
) -> pd.DataFrame:
    """Synthetic dataframe generator.

    Realises 0-variation outcome mixed with linear x dependency.

    Args:
        data_size:
            Length of output Dataframe
        dependent_division:
            If True - the returned Dataframe contains a division into
            groups dependent on the features.
            If False - the returned Dataframe contains a division into
            groups independent on the features.
        random_state:
            If specified - defines numpy random seed. Defaults to None.

    Returns:
        Synthetic dataframe also containing fictional information about
        features and lagged features.
    """
    if random_state is not None:
        np.random.seed(random_state)

    mean_x_feature = np.random.uniform(0, 5, size=data_size)

    x_feature_lag_1 = np.random.normal(mean_x_feature, 2)
    x_feature = np.random.normal(mean_x_feature, 2)

    treatment = sigmoid_division(x_feature_lag_1, dependent_division)

    target_lag_1 = 200 + x_feature_lag_1 * 100
    target_factual = 200 + x_feature * 100 + treatment * 10

    df = pd.DataFrame(
        dict(
            X_lag_1=x_feature_lag_1,
            Target_lag_1=target_lag_1,
            X=x_feature,
            Treatment=treatment,
            Target=target_factual,
        )
    )
    return df



from __future__ import annotations

from copy import deepcopy
from typing import Any, Sequence

import numpy as np

from ..dataset import (
    ABCRole,
    AdditionalFeatureRole,
    Dataset,
    ExperimentData,
    FeatureRole,
    GroupingRole,
    TargetRole,
)
from ..executor import Calculator
from ..extensions.scipy_linalg import CholeskyExtension, InverseExtension
from ..utils import ExperimentDataEnum, NotSuitableFieldError
from ..utils.adapter import Adapter


class MahalanobisDistance(Calculator):
    def __init__(
        self,
        grouping_role: ABCRole | None = None,
        key: Any = "",
        weights: dict[str, float] | None = None,
    ):
        super().__init__(key=key)
        self.grouping_role = grouping_role or GroupingRole()
        self.weights = weights

    @classmethod
    def _execute_inner_function(
        cls,
        grouping_data,
        target_fields: list[str] | None = None,
        **kwargs,
    ) -> dict:
        result = {}
        for i in range(1, len(grouping_data)):
            result.update(
                cls._inner_function(
                    data=(
                        grouping_data[0][1][target_fields]
                        if target_fields
                        else grouping_data[0][1]
                    ),
                    test_data=(
                        grouping_data[i][1][target_fields]
                        if target_fields
                        else grouping_data[i][1]
                    ),
                    **kwargs,
                )
            )
        return result

    def _set_value(
        self, data: ExperimentData, value: dict | None = None, key: Any = None
    ) -> ExperimentData:
        for key, value_ in value.items():
            data = data.set_value(
                ExperimentDataEnum.groups,
                self.id,
                value_,
                key=key,
            )
        return data

    def _get_fields(self, data: ExperimentData):
        group_field = data.field_search(self.grouping_role)
        target_fields = data.field_search(
            [FeatureRole(), AdditionalFeatureRole()], search_types=self.search_types
        )
        return group_field, target_fields

    @property
    def search_types(self) -> list[type] | None:
        return [int, float]

    @classmethod
    def _inner_function(
        cls,
        data: Dataset,
        test_data: Dataset | None = None,
        weights: dict[str, float] | None = None,
        **kwargs,
    ):
        test_data = cls._check_test_data(test_data)
        cov = (data.cov() + test_data.cov()) / 2 if test_data else data.cov()
        cholesky = CholeskyExtension().calc(cov)
        mahalanobis_transform = InverseExtension().calc(cholesky)
        if weights is not None:
            features = data.columns
            w_list = np.array(
                [weights[col] if col in weights.keys() else 1 for col in features]
            )
            w_matrix = np.sqrt(np.diag(w_list / w_list.sum()))
            mahalanobis_transform = mahalanobis_transform.dot(w_matrix)
        y_control = data.dot(mahalanobis_transform.transpose())
        if test_data:
            y_test = test_data.dot(mahalanobis_transform.transpose())
            return {"control": y_control, "test": y_test}
        return {"control": y_control}

    @classmethod
    def calc(
        cls,
        data: Dataset,
        group_field: Sequence[str] | str | None = None,
        grouping_data: list[tuple[str, Dataset]] | None = None,
        target_fields: str | list[str] | None = None,
        weights: dict[str, float] | None = None,
        **kwargs,
    ) -> dict:
        group_field = Adapter.to_list(group_field)

        if grouping_data is None:
            grouping_data = data.groupby(group_field)
        if len(grouping_data) > 1:
            grouping_data[0][1].tmp_roles = data.tmp_roles
        else:
            raise NotSuitableFieldError(group_field, "Grouping")
        return cls._execute_inner_function(
            grouping_data,
            target_fields=target_fields,
            old_data=data,
            weights=weights,
            **kwargs,
        )

    def execute(self, data: ExperimentData) -> ExperimentData:
        group_field, target_fields = self._get_fields(data=data)
        self.key = str(
            target_fields[0] if len(target_fields) == 1 else (target_fields or "")
        )
        if (
            not target_fields and data.ds.tmp_roles
        ):  # if the column is not suitable for the test, then the target will be empty, but if there is a role tempo, then this is normal behavior
            return data
        if group_field[0] in data.groups:  # TODO: to recheck if this is a correct check
            grouping_data = list(data.groups[group_field[0]].items())
        else:
            grouping_data = None
        t_data = deepcopy(data.ds)
        for field in target_fields:
            if field not in t_data.columns:
                t_data = t_data.add_column(
                    data.additional_fields[field],
                    role={field: TargetRole()},
                )
        compare_result = self.calc(
            data=t_data,
            group_field=group_field,
            target_fields=target_fields,
            grouping_data=grouping_data,
            weights=self.weights or None,
        )
        return self._set_value(data, compare_result)



from __future__ import annotations

from typing import Any, ClassVar

import numpy as np

from ..comparators import Chi2Test, KSTest, TTest
from ..dataset import Dataset, ExperimentData, StatisticRole
from ..executor import Executor
from ..experiments.base_complex import IfParamsExperiment, ParamsExperiment
from ..reporters.aa import OneAADictReporter
from ..splitters import AASplitter, AASplitterWithStratification
from ..utils import ID_SPLIT_SYMBOL, BackendsEnum, ExperimentDataEnum


class OneAAStatAnalyzer(Executor):
    def _set_value(self, data: ExperimentData, value, key=None) -> ExperimentData:
        return data.set_value(ExperimentDataEnum.analysis_tables, self.id, value)

    def execute(self, data: ExperimentData) -> ExperimentData:
        analysis_tests: list[type] = [TTest, KSTest, Chi2Test]
        executor_ids = data.get_ids(
            analysis_tests, searched_space=ExperimentDataEnum.analysis_tables
        )
        # num_groups = len(data.groups[data.ds.search_columns(TreatmentRole())[0]]) - 1
        # groups = list(data.groups[data.ds.search_columns(TreatmentRole())[0]].items())
        # multitest_pvalues = Dataset.create_empty()
        # analysis_data = {}

        analysis_data: dict[str, float] = {}
        for class_, spaces in executor_ids.items():
            analysis_ids = spaces.get("analysis_tables", [])
            if len(analysis_ids) > 0:
                if len(analysis_ids) > 1:
                    t_data = data.analysis_tables[analysis_ids[0]].append(
                        [data.analysis_tables[k] for k in analysis_ids[1:]]
                    )
                else:
                    t_data = data.analysis_tables[analysis_ids[0]]
                # t_data.data.index = analysis_ids
                for field in ["p-value", "pass"]:
                    analysis_data[f"mean {class_} {field}"] = t_data[field].mean()
        analysis_data["mean test score"] = 0
        sum_weight = 0
        analysis_data = {
            key: (0 if np.isnan(value) else value)
            for key, value in analysis_data.items()
        }
        if (
            "mean TTest p-value" in analysis_data
            and "mean KSTest p-value" in analysis_data
        ):
            analysis_data["mean test score"] = (
                analysis_data["mean TTest p-value"]
                + 2 * analysis_data["mean KSTest p-value"]
            )
            sum_weight += 3
        if "mean Chi2Test p-value" in analysis_data:
            analysis_data["mean test score"] += (
                2 * analysis_data["mean Chi2Test p-value"]
            )
            sum_weight += 2
        if sum_weight:
            analysis_data["mean test score"] /= sum_weight

        analysis_dataset = Dataset.from_dict(
            [analysis_data],
            {field: StatisticRole() for field in analysis_data},
            BackendsEnum.pandas,
        )

        return self._set_value(data, analysis_dataset)


class AAScoreAnalyzer(Executor):
    AA_SPLITER_CLASS_MAPPING: ClassVar[dict] = {
        class_.__name__: class_ for class_ in [AASplitter, AASplitterWithStratification]
    }

    # TODO: rename alpha
    def __init__(self, alpha: float = 0.05, key: str = ""):
        super().__init__(key=key)
        self.alpha = alpha
        self.__feature_weights = {}
        self.threshold = 1 - (self.alpha * 1.2)

    def _set_value(
        self, data: ExperimentData, value: Any, key: Any = None
    ) -> ExperimentData:
        return data.set_value(
            ExperimentDataEnum.analysis_tables,
            executor_id=self.id,
            key=self.key,
            value=value,
        )

    def _analyze_aa_score(
        self, data: ExperimentData, score_table: Dataset
    ) -> ExperimentData:
        search_flag = f"{ID_SPLIT_SYMBOL}pass{ID_SPLIT_SYMBOL}"
        self.__feature_weights = {
            column: 1 - abs(self.alpha - score_table.loc[:, column].mean())
            for column in score_table.columns
            if search_flag in column
        }
        aa_scores = {
            class_.replace(f"{ID_SPLIT_SYMBOL}pass", ""): value
            for class_, value in self.__feature_weights.items()
        }
        aa_passed = {
            class_: value >= self.threshold for class_, value in aa_scores.items()
        }
        result = Dataset.from_dict({"score": aa_scores, "pass": aa_passed}, roles={})
        self.key = "aa score"
        return self._set_value(data, result)

    def build_splitter_from_id(self, splitter_id: str):
        splitter_class = self.AA_SPLITER_CLASS_MAPPING.get(
            splitter_id[: splitter_id.find(ID_SPLIT_SYMBOL)]
        )
        if splitter_class is None:
            raise ValueError(f"{splitter_id} is not a valid splitter id")
        return splitter_class.build_from_id(splitter_id)

    def _get_best_split(
        self,
        data: ExperimentData,
        score_table: Dataset,
        if_param_scores: Dataset | None = None,
    ) -> dict[str, Any]:
        # TODO: add split_scores in ExperimentData
        if if_param_scores is None:
            if len(self.__feature_weights) < 1:
                best_index = 0
            else:
                aa_split_scores = score_table.apply(
                    lambda x: (
                        (
                            (
                                (
                                    sum(
                                        x[
                                            key.replace(
                                                f"{ID_SPLIT_SYMBOL}pass{ID_SPLIT_SYMBOL}",
                                                f"{ID_SPLIT_SYMBOL}p-value{ID_SPLIT_SYMBOL}",
                                            )
                                        ]
                                        * value
                                        for key, value in self.__feature_weights.items()
                                        if (isinstance(value, float) and value > 0)
                                        and (
                                            key.replace(
                                                f"{ID_SPLIT_SYMBOL}pass{ID_SPLIT_SYMBOL}",
                                                f"{ID_SPLIT_SYMBOL}p-value{ID_SPLIT_SYMBOL}",
                                            )
                                            in x["splitter_id"]
                                        )
                                    )
                                    / len(self.__feature_weights)
                                )
                                * 2
                            )
                            / 3
                        )
                        + x["mean test score"] / 3
                    ),
                    axis=1,
                    role={"aa split score": StatisticRole()},
                )
                best_index = aa_split_scores.idxmax()
            best_split_id = score_table.loc[best_index, "splitter_id"].get_values(0, 0)
            score_dict = score_table.loc[best_index, :].transpose().to_records()[0]
        else:
            best_index = 0
            best_split_id = score_table.loc[best_index, "splitter_id"].get_values(0, 0)
            score_dict = if_param_scores.loc[best_index, :].transpose().to_records()[0]
        best_score_stat = OneAADictReporter.convert_flat_dataset(score_dict)
        self.key = "best split statistics"
        result = self._set_value(data, best_score_stat)
        return {"best_split_id": best_split_id, "data": result}

    def _set_best_split(
        self,
        data: ExperimentData,
        best_splitter_id: str,
    ) -> ExperimentData:
        self.key = "best splitter"
        result = data.set_value(
            ExperimentDataEnum.variables, self.id, best_splitter_id, self.key
        )
        best_splitter = self.build_splitter_from_id(best_splitter_id)
        best_splitter.save_groups = False
        best_splitter.constant_key = False
        best_splitter.key = "best"
        result = best_splitter.execute(result)
        return result

    def _analyze_best_split(
        self,
        data: ExperimentData,
        score_table: Dataset,
        if_param_scores: Dataset | None = None,
    ) -> ExperimentData:
        best_split = self._get_best_split(data, score_table, if_param_scores)
        return self._set_best_split(best_split["data"], best_split["best_split_id"])

    def execute(self, data: ExperimentData) -> ExperimentData:
        param_experiment_id = data.get_one_id(
            ParamsExperiment, ExperimentDataEnum.analysis_tables, "AATest"
        )
        ifparam_experiment_id = data.get_ids(
            IfParamsExperiment,
            ExperimentDataEnum.analysis_tables,
        )
        score_table = data.analysis_tables[param_experiment_id]
        score_table = score_table.dropna(axis=1, how="all")
        if_param_scores = (
            None
            if len(ifparam_experiment_id["IfParamsExperiment"]["analysis_tables"]) == 0
            else data.analysis_tables[
                ifparam_experiment_id["IfParamsExperiment"]["analysis_tables"][0]
            ]
        )
        data = self._analyze_aa_score(data, score_table)
        return self._analyze_best_split(data, score_table, if_param_scores)



from abc import abstractmethod

from ..dataset import Dataset, ExperimentData
from ..executor import Calculator
from ..utils import AbstractMethodError


class Transformer(Calculator):
    @property
    def _is_transformer(self):
        return True

    @staticmethod
    @abstractmethod
    def _inner_function(data: Dataset, **kwargs) -> Dataset:
        raise AbstractMethodError

    @classmethod
    def calc(cls, data: Dataset, **kwargs):
        return cls._inner_function(data, **kwargs)

    def execute(self, data: ExperimentData) -> ExperimentData:
        data = data.copy(data=self.calc(data=data.ds))
        return data



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Sequence

from ..dataset import (
    ABCRole,
    AdditionalMatchingRole,
    Dataset,
    ExperimentData,
    FeatureRole,
    GroupingRole,
    TargetRole,
)
from ..utils import (
    ID_SPLIT_SYMBOL,
    AbstractMethodError,
    ExperimentDataEnum,
    NotSuitableFieldError,
    SetParamsDictTypes,
)
from ..utils.adapter import Adapter


class Executor(ABC):
    def __init__(
        self,
        key: Any = "",
        **calc_kwargs,
    ):
        self._id: str = ""
        self._params_hash = ""

        self.key: Any = key
        self._generate_id()
        self.calc_kwargs = calc_kwargs

    def check_and_setattr(self, params: dict[str, Any]):
        for key, value in params.items():
            if key in self.__dir__():
                setattr(self, key, value)

    def _generate_params_hash(self):
        self._params_hash = ""

    def _generate_id(self):
        self._generate_params_hash()
        self._id = ID_SPLIT_SYMBOL.join(
            [
                self.__class__.__name__,
                self._params_hash.replace(ID_SPLIT_SYMBOL, "|"),
                str(self._key).replace(ID_SPLIT_SYMBOL, "|"),
            ]
        )

    def set_params(self, params: SetParamsDictTypes) -> None:
        if isinstance(next(iter(params)), str):
            self.check_and_setattr(params)
        elif isinstance(next(iter(params)), type):
            for executor_class, class_params in params.items():
                if isinstance(self, executor_class):
                    self.check_and_setattr(class_params)
        else:
            raise ValueError(
                "params must be a dict of str to dict or a dict of class to dict"
            )
        self._generate_id()

    def init_from_hash(self, hash: str) -> None:
        self._params_hash = hash
        self._generate_id()

    @classmethod
    def build_from_id(cls, executor_id: str):
        splitted_id = executor_id.split(ID_SPLIT_SYMBOL)
        if splitted_id[0] != cls.__name__:
            raise ValueError(f"{executor_id} is not a valid {cls.__name__} id")
        result = cls()
        result.init_from_hash(splitted_id[1])
        return result

    @property
    def id(self) -> str:
        return self._id

    @property
    def key(self) -> Any:
        return self._key

    @key.setter
    def key(self, value: Any):
        self._key = value
        self._generate_id()

    @property
    def params_hash(self) -> str:
        return self._params_hash

    @property
    def id_for_name(self) -> str:
        return self.id.replace(ID_SPLIT_SYMBOL, "_")

    @property
    def _is_transformer(self) -> bool:
        return False

    def _set_value(
        self, data: ExperimentData, value: Any, key: Any = None
    ) -> ExperimentData:
        # defined in order to avoid  unnecessary redefinition in classes like transformer
        return data

    @abstractmethod
    def execute(self, data: ExperimentData) -> ExperimentData:
        raise AbstractMethodError


class Calculator(Executor, ABC):
    @classmethod
    def calc(cls, data: Dataset, **kwargs):
        return cls._inner_function(data, **kwargs)

    @staticmethod
    @abstractmethod
    def _inner_function(data: Dataset, **kwargs) -> Any:
        raise AbstractMethodError

    @property
    def search_types(self):
        raise AbstractMethodError

    @staticmethod
    def _check_test_data(
        test_data: Dataset | None = None,
    ) -> Dataset:  # TODO to move away from Calculator. Where to?
        if test_data is None:
            raise ValueError("test_data is needed for comparison")
        return test_data


class MLExecutor(Calculator, ABC):
    def __init__(
        self,
        grouping_role: ABCRole | None = None,
        target_role: ABCRole | None = None,
        key: Any = "",
    ):
        self.target_role = target_role or TargetRole()
        super().__init__(key=key)
        self.grouping_role = grouping_role or GroupingRole()

    def _get_fields(self, data: ExperimentData):
        group_field = data.field_search(self.grouping_role)
        target_field = data.field_search(
            self.target_role, search_types=self.search_types
        )
        return group_field, target_field

    @abstractmethod
    def fit(self, X: Dataset, Y: Dataset | None = None) -> MLExecutor:
        raise NotImplementedError

    @abstractmethod
    def predict(self, X: Dataset) -> Dataset:
        raise NotImplementedError

    def score(self, X: Dataset, Y: Dataset) -> float:
        raise NotImplementedError

    @property
    def search_types(self):
        return [int, float]

    @classmethod
    @abstractmethod
    def _inner_function(
        cls,
        data: Dataset,
        test_data: Dataset | None = None,
        target_data: Dataset | None = None,
        **kwargs,
    ) -> Any:
        raise AbstractMethodError

    @classmethod
    def _execute_inner_function(
        cls,
        grouping_data,
        target_field: str | None = None,
        **kwargs,
    ) -> Any:
        if target_field:
            return cls._inner_function(
                data=grouping_data[0][1].drop(target_field),
                target_data=grouping_data[0][1][target_field],
                test_data=grouping_data[1][1].drop(target_field),
                **kwargs,
            )
        return cls._inner_function(
            data=grouping_data[0][1],
            test_data=grouping_data[1][1],
            **kwargs,
        )

    def _set_value(
        self, data: ExperimentData, value: Any, key: Any = None
    ) -> ExperimentData:
        for i in range(value.shape[1]):
            data.set_value(
                ExperimentDataEnum.additional_fields,
                f"{self.id}{ID_SPLIT_SYMBOL}{i}",
                value=value.iloc[:, i],
                key=key,
                role=AdditionalMatchingRole(),
            )
        return data

    @classmethod
    def calc(
        cls,
        data: Dataset,
        group_field: Sequence[str] | str | None = None,
        grouping_data: list[tuple[str, Dataset]] | None = None,
        target_field: str | list[str] | None = None,
        features_fields: str | list[str] | None = None,
        **kwargs,
    ) -> Dataset:
        group_field = Adapter.to_list(group_field)
        features_fields = Adapter.to_list(features_fields)
        if grouping_data is None:
            grouping_data = data.groupby(group_field, fields_list=features_fields)
        if len(grouping_data) > 1:
            grouping_data[0][1].tmp_roles = data.tmp_roles
        else:
            raise NotSuitableFieldError(group_field, "Grouping")
        result = cls._execute_inner_function(
            grouping_data, target_field=target_field, **kwargs
        )
        return result

    def execute(self, data: ExperimentData) -> ExperimentData:
        group_field, target_fields = self._get_fields(data=data)
        features_fields = data.ds.search_columns(
            FeatureRole(), search_types=self.search_types
        )
        self.key = str(
            target_fields[0] if len(target_fields) == 1 else (target_fields or "")
        )
        if (
            not target_fields and data.ds.tmp_roles
        ):  # if the column is not suitable for the test, then the target will be empty, but if there is a role tempo, then this is normal behavior
            return data
        if group_field[0] in data.groups:  # TODO: to recheck if this is a correct check
            grouping_data = list(data.groups[group_field[0]].items())
        else:
            grouping_data = None
        compare_result = self.calc(
            data=data.ds,
            group_field=group_field,
            grouping_data=grouping_data,
            target_fields=target_fields,
            features_fields=features_fields,
        )
        # TODO: add roles to compare_result
        return self._set_value(data, compare_result)


class IfExecutor(Executor, ABC):
    def __init__(
        self,
        if_executor: Executor | None = None,
        else_executor: Executor | None = None,
        key: Any = "",
    ):
        self.if_executor = if_executor
        self.else_executor = else_executor
        super().__init__(key)

    @abstractmethod
    def check_rule(self, data, **kwargs) -> bool:
        raise AbstractMethodError

    def _set_value(
        self, data: ExperimentData, value: Any, key: Any = None
    ) -> ExperimentData:
        return data.set_value(
            ExperimentDataEnum.variables, self.id, value, key="response"
        )

    def execute(self, data: ExperimentData) -> ExperimentData:
        if self.check_rule(data):
            return (
                self.if_executor.execute(data)
                if self.if_executor is not None
                else self._set_value(data, True)
            )
        return (
            self.else_executor.execute(data)
            if self.else_executor is not None
            else self._set_value(data, False)
        )



from __future__ import annotations

from functools import wraps
from typing import Any, Callable, cast

from hypex.utils import DecoratedType, DocstringInheritDecorator


def inherit_docstring_from(
    source: Callable[..., Any] | property,
) -> DocstringInheritDecorator:
    """A decorator to inherit the docstring from another function or property.

    This decorator can be applied to both callable objects and properties. It copies the docstring
    from the source object to the decorated object if the latter does not already have a docstring.

    Args:
        source: The object from which the docstring will be inherited.
            This should be either a callable or a property that has
            a well-defined __doc__ attribute.

    Returns:
        A decorator that when applied to a function or property,
        sets its __doc__ attribute to that of the source.

    Raises:
        TypeError: If the object to be decorated is neither a callable nor a property.

    Example:
        Using with property::

            class SomeClass:
                @property
                @inherit_docstring_from(pd.DataFrame.iloc)
                def iloc(self):
                    return self._data.iloc

        Using with method::

            @inherit_docstring_from(pd.DataFrame.mean)
            def mean(self):
                return self._data.mean()
    """

    def decorator(obj: DecoratedType) -> DecoratedType:
        """
        Apply the inherited docstring to a given function or property.
        This function acts as a decorator within 'inherit_docstring_from', applying the docstring
        from the 'source' object to the 'obj'. If 'obj' is a property, it modifies the property to include
        the source's docstring. If 'obj' is a callable, it wraps the callable in a function that preserves
        the original callable's functionality and metadata but updates the docstring.
        Args:
            obj (DecoratedType): The function or property to which the docstring will be applied.
                                 It must be either a callable or a property.
        Returns:
            DecoratedType: The original object with the updated docstring. If the object is a property,
                           it returns a new property object with the inherited docstring. If it's a callable,
                           it returns the wrapped callable with the updated docstring.
        Raises:
            TypeError: If 'obj' is neither a callable nor a property.
        """
        if isinstance(obj, property):
            doc = getattr(source, "__doc__", "No documentation provided.")
            return property(obj.fget, obj.fset, obj.fdel, doc)
        elif callable(obj):

            @wraps(obj)
            def wrapper(*args, **kwargs) -> Any:
                return obj(*args, **kwargs)

            wrapper.__doc__ = getattr(source, "__doc__", "No documentation provided.")
            return cast(DecoratedType, wrapper)
        else:
            raise TypeError(
                "The decorator can only be applied to callables or properties."
            )

    return decorator



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Literal

from ..dataset import ABCRole, Dataset
from ..dataset.backends import PandasDataset
from ..dataset.dataset import DatasetAdapter
from ..utils.errors import AbstractMethodError


class Extension(ABC):
    def __init__(self):
        self.BACKEND_MAPPING = {
            PandasDataset: self._calc_pandas,
        }

    @abstractmethod
    def _calc_pandas(self, data: Dataset, **kwargs):
        raise AbstractMethodError

    def calc(self, data: Dataset, **kwargs):
        return self.BACKEND_MAPPING[type(data.backend)](data=data, **kwargs)

    @staticmethod
    def result_to_dataset(result: Any, roles: ABCRole | dict[str, ABCRole]) -> Dataset:
        return DatasetAdapter.to_dataset(result, roles=roles)


class CompareExtension(Extension, ABC):
    def calc(self, data: Dataset, other: Dataset | None = None, **kwargs):
        return super().calc(data=data, other=other, **kwargs)


class MLExtension(Extension):
    #   TODO: add model
    def _calc_pandas(
        self,
        data: Dataset,
        mode: Literal["auto", "fit", "predict"] | None = None,
        **kwargs,
    ):
        if mode in ["auto", "fit"]:
            return self.fit(data, **kwargs)
        return self.predict(data, **kwargs)

    @abstractmethod
    def fit(self, X, Y=None, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def predict(self, X, **kwargs):
        raise NotImplementedError

    def calc(
        self,
        data: Dataset,
        **kwargs,
    ):
        return super().calc(data=data, **kwargs)



from __future__ import annotations

from typing import Literal

from .analyzers.matching import MatchingAnalyzer
from .comparators import Chi2Test, KSTest, TTest
from .comparators.distances import MahalanobisDistance
from .dataset import AdditionalMatchingRole, FeatureRole, TargetRole, TreatmentRole
from .encoders.encoders import DummyEncoder
from .executor import Executor
from .experiments import GroupExperiment
from .experiments.base import Experiment, OnRoleExperiment
from .ml.faiss import FaissNearestNeighbors
from .operators.operators import Bias, MatchingMetrics
from .reporters.matching import MatchingDatasetReporter
from .transformers import TypeCaster
from .ui.base import ExperimentShell
from .ui.matching import MatchingOutput


class Matching(ExperimentShell):
    """A class for performing matching analysis with configurable distance metrics and quality tests.

    This class provides functionality to identify groups that are similar to each other
    using various distance metrics and quality assessment methods.

    Args:
        group_match (bool, optional): Whether to perform group matching. Defaults to False.
        distance (Literal["mahalanobis", "l2"], optional): Distance metric to use for matching.
            Options are "mahalanobis" or "l2". Defaults to "mahalanobis".
        metric (Literal["atc", "att", "ate"], optional): Type of treatment effect to estimate.
            "atc" = average treatment effect on controls
            "att" = average treatment effect on treated
            "ate" = average treatment effect
            Defaults to "ate".
        bias_estimation (bool, optional): Whether to estimate bias. Defaults to True.
        quality_tests (Union[str, List[str]], optional): Quality tests to perform.
            Options are "smd", "psi", "ks-test", "repeats", "t-test", or "auto".
            Can be a single test or list of tests. Defaults to "auto".
        faiss_mode (Literal["base", "fast", "auto"], optional): Faiss mode to use for matching.
            Options are "base", "fast", or "auto". Defaults to "auto".
        n_neighbors (int, optional): Number of neighbors to use for matching. Defaults to 1.

    Examples
    --------
    .. code-block:: python

        # Basic matching with default settings
        matching = Matching()
        results = matching.execute(data)

        # Matching with L2 distance and specific quality tests
        matching = Matching(
            distance="l2",
            quality_tests=["t-test", "ks-test"]
        )
        results = matching.execute(data)

        # Group matching with ATT estimation
        matching = Matching(
            group_match=True,
            metric="att",
            bias_estimation=True
        )
        results = matching.execute(data)
    """

    @staticmethod
    def _make_experiment(
        group_match: bool = False,
        distance: Literal["mahalanobis", "l2"] = "mahalanobis",
        metric: Literal["atc", "att", "ate"] = "ate",
        bias_estimation: bool = True,
        quality_tests: (
            Literal["smd", "psi", "ks-test", "repeats", "t-test", "chi2-test", "auto"]
            | list[
                Literal[
                    "smd", "psi", "ks-test", "repeats", "t-test", "chi2-test", "auto"
                ]
            ]
        ) = "auto",
        faiss_mode: Literal["base", "fast", "auto"] = "auto",
        n_neighbors: int = 1,
        weights: dict[str, float] | None = None,
        encode_categories: bool = True,
    ) -> Experiment:
        """Creates an experiment configuration with specified matching parameters.

        Args:
            group_match (bool, optional): Whether to perform group matching. Defaults to False.
        distance (Literal["mahalanobis", "l2"], optional): Distance metric to use for matching.
            Options are "mahalanobis" or "l2". Defaults to "mahalanobis".
        metric (Literal["atc", "att", "ate"], optional): Type of treatment effect to estimate.
            "atc" = average treatment effect on controls
            "att" = average treatment effect on treated
            "ate" = average treatment effect
            Defaults to "ate".
        bias_estimation (bool, optional): Whether to estimate bias. Defaults to True.
        quality_tests (Union[str, List[str]], optional): Quality tests to perform.
            Options are "ks-test", "t-test", or "auto".
            Can be a single test or list of tests. Defaults to "auto", which performs all tests.
        faiss_mode (Literal["base", "fast", "auto"], optional): Faiss mode to use for matching.
            Options are "base", "fast", or "auto". Defaults to "auto".
        n_neighbors (int, optional): Number of neighbors to use for matching. Defaults to 1.

        Returns:
            Experiment: Configured experiment object with specified matching parameters.

        Examples
        --------
        .. code-block:: python

            exp = Matching._make_experiment(
                distance="l2",
                metric="att",
                quality_tests=["t-test"]
            )
        """
        distance_mapping = {
            "mahalanobis": MahalanobisDistance(
                grouping_role=TreatmentRole(), weights=weights
            ),
            # "l2": L2Distance(grouping_role=TreatmentRole(), weights=weights),
        }
        test_mapping = {
            "t-test": TTest(
                compare_by="matched_pairs",
                grouping_role=TreatmentRole(),
                baseline_role=AdditionalMatchingRole(),
            ),
            # "psi": PSI(grouping_role=TreatmentRole(), compare_by="groups"),
            "ks-test": KSTest(
                grouping_role=TreatmentRole(),
                compare_by="matched_pairs",
                baseline_role=AdditionalMatchingRole(),
            ),
            "chi2-test": Chi2Test(
                grouping_role=TreatmentRole(),
                compare_by="matched_pairs",
                baseline_role=AdditionalMatchingRole(),
            ),
        }
        two_sides = metric == "ate"
        test_pairs = metric == "atc"
        executors: list[Executor] = [
            TypeCaster(
                dtype={int: float},
                roles=[FeatureRole(), TargetRole()],
            ),
            FaissNearestNeighbors(
                grouping_role=TreatmentRole(),
                two_sides=two_sides,
                test_pairs=test_pairs,
                faiss_mode=faiss_mode,
                n_neighbors=n_neighbors,
            ),
        ]
        if bias_estimation:
            executors += [
                Bias(grouping_role=TreatmentRole(), target_roles=[TargetRole()]),
            ]
        executors += [
            MatchingMetrics(
                grouping_role=TreatmentRole(),
                target_roles=[TargetRole()],
                metric=metric,
                n_neighbors=n_neighbors,
            ),
            MatchingAnalyzer(),
        ]
        if quality_tests == "auto":
            executors += [
                OnRoleExperiment(
                    executors=list(test_mapping.values()),
                    role=FeatureRole(),
                )
            ]
        else:
            # warnings.warn("Now quality tests aren't supported yet")
            executors += [
                OnRoleExperiment(
                    executors=[test_mapping[test] for test in quality_tests],
                    role=FeatureRole(),
                )
            ]
        executors = (
            executors if distance == "l2" else [distance_mapping[distance], *executors]
        )
        executors = executors if not encode_categories else [DummyEncoder(), *executors]
        return (
            Experiment(executors=executors)
            if not group_match
            else GroupExperiment(
                executors=executors,
                reporter=MatchingDatasetReporter(),
            )
        )

    def __init__(
        self,
        group_match: bool = False,
        distance: Literal["mahalanobis", "l2"] = "mahalanobis",
        # metric: Literal["atc", "att", "ate"] = "ate",
        bias_estimation: bool = True,
        quality_tests: (
            Literal["smd", "psi", "ks-test", "repeats", "t-test", "chi2-test", "auto"]
            | list[
                Literal[
                    "smd", "psi", "ks-test", "repeats", "t-test", "chi2-test", "auto"
                ]
            ]
        ) = "auto",
        faiss_mode: Literal["base", "fast", "auto"] = "auto",
        n_neighbors: int = 1,
        weights: dict[str, float] | None = None,
        encode_categories: bool = True,
    ):
        metric = "ate"
        super().__init__(
            experiment=self._make_experiment(
                group_match,
                distance,
                metric,
                bias_estimation,
                quality_tests,
                faiss_mode,
                n_neighbors,
                weights,
                encode_categories,
            ),
            output=MatchingOutput(GroupExperiment if group_match else MatchingAnalyzer),
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Callable, Iterable, Literal, Sequence, Sized

from ...utils import AbstractMethodError, FromDictTypes


class DatasetBackendNavigation(ABC):
    @property
    def name(self) -> str:
        return str(self.__class__.__name__).lower().replace("backend", "")

    @property
    @abstractmethod
    def index(self):
        raise AbstractMethodError

    @property
    @abstractmethod
    def columns(self):
        raise AbstractMethodError

    @abstractmethod
    def from_dict(self, data: FromDictTypes, index: Iterable | Sized | None = None):
        raise AbstractMethodError

    @abstractmethod
    def to_dict(self) -> dict[str, Any]:
        raise AbstractMethodError

    def to_records(self) -> list[dict]:
        raise AbstractMethodError

    @abstractmethod
    def __getitem__(self, item) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __len__(self) -> int:
        raise AbstractMethodError

    # Comparison methods:
    @abstractmethod
    def __eq__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __ne__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __le__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __lt__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __ge__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __gt__(self, other) -> Any:
        raise AbstractMethodError

    # Unary methods:
    @abstractmethod
    def __pos__(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __neg__(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __abs__(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __invert__(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __round__(self, ndigits: int = 0) -> Any:
        raise AbstractMethodError

    # Binary methods:
    @abstractmethod
    def __add__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __sub__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __mul__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __floordiv__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __div__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __truediv__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __mod__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __pow__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __and__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __or__(self, other) -> Any:
        raise AbstractMethodError

    # Right arithmetic methods:
    @abstractmethod
    def __radd__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __rsub__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __rmul__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __rfloordiv__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __rdiv__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __rtruediv__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __rmod__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __rpow__(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def __repr__(self):
        raise AbstractMethodError

    @abstractmethod
    def _repr_html_(self):
        raise AbstractMethodError

    @abstractmethod
    def create_empty(
        self,
        index: Iterable | None = None,
        columns: Iterable[str] | None = None,
    ):
        raise AbstractMethodError

    @abstractmethod
    def _get_column_index(
        self, column_name: Sequence[str] | str
    ) -> int | Sequence[int]:
        raise AbstractMethodError

    @abstractmethod
    def get_column_type(self, column_name: str) -> type | None:
        raise AbstractMethodError

    @abstractmethod
    def astype(
        self, dtype: dict[str, type], errors: Literal["raise", "ignore"] = "raise"
    ) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def update_column_type(self, column_name: str, type_name: type):
        raise AbstractMethodError

    @abstractmethod
    def add_column(
        self,
        data: Sequence,
        name: str,
        index: Sequence | None = None,
    ):
        raise AbstractMethodError

    @abstractmethod
    def append(self, other, index: bool = False) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def loc(self, values: Iterable) -> Iterable:
        raise AbstractMethodError

    @abstractmethod
    def iloc(self, values: Iterable) -> Iterable:
        raise AbstractMethodError


class DatasetBackendCalc(DatasetBackendNavigation, ABC):
    @abstractmethod
    def mean(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def mode(self, numeric_only: bool = False, dropna: bool = True) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def var(
        self, skipna: bool = True, ddof: int = 1, numeric_only: bool = False
    ) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def max(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def idxmax(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def min(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def count(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def sum(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def log(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def agg(self, func) -> Any:
        raise AbstractMethodError

    def get(
        self,
        key,
        default=None,
    ) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def take(
        self,
        indices: int | list[int],
        axis: Literal["index", "columns", "rows"] | int = 0,
    ) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def get_values(
        self,
        row: str | None = None,
        column: str | None = None,
    ) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def iget_values(
        self,
        row: int | None = None,
        column: int | None = None,
    ) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def apply(self, func: Callable, **kwargs) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def map(self, func: Callable, **kwargs) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def is_empty(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def unique(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def nunique(self, dropna: bool = True) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def isin(self, values: Iterable) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def groupby(self, by: str | Iterable[str], **kwargs) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def sort_index(self, **kwargs) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def sort_values(self, by: str | list[str], ascending: bool = True, **kwargs) -> Any:
        raise AbstractMethodError

    def std(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def coefficient_of_variation(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def corr(self, method="pearson", numeric_only=False) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def value_counts(
        self,
        normalize: bool = False,
        sort: bool = True,
        ascending: bool = False,
        dropna: bool = True,
    ) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def na_counts(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def dropna(
        self,
        how: Literal["any", "all"] = "any",
        subset: str | Iterable[str] | None = None,
        axis: Literal["index", "rows", "columns"] | int = 0,
    ) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def isna(self) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def quantile(self, q: float = 0.5) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def select_dtypes(
        self, include: Any | None = None, exclude: Any | None = None
    ) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def merge(
        self,
        right: Any,
        on: str | None = None,
        left_on: str | None = None,
        right_on: str | None = None,
        left_index: bool = False,
        right_index: bool = False,
        suffixes: tuple[str, str] = ("_x", "_y"),
        how: Literal["left", "right", "inner", "outer", "cross"] = "inner",
    ) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def drop(
        self,
        labels: str | None = None,
        axis: int | None = None,
        columns: str | Iterable[str] | None = None,
    ) -> Any:
        raise AbstractMethodError

    def filter(
        self,
        items: list | None = None,
        like: str | None = None,
        regex: str | None = None,
        axis: int = 0,
    ) -> Any:
        return AbstractMethodError

    def fillna(self, values, method, **kwargs) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def dot(self, other) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def transpose(self, names) -> Any:
        raise AbstractMethodError

    @abstractmethod
    def rename(self, columns: dict[str, str]) -> Any:
        return AbstractMethodError

    @abstractmethod
    def replace(
        self, to_replace: Any = None, value: Any = None, regex: bool = False
    ) -> Any:
        raise AbstractMethodError
